#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
PyQt graphical interface to gaitutils

@author: Jussi (jnu@iki.fi)
"""

from __future__ import print_function
from __future__ import division

from builtins import str
from PyQt5 import QtGui, QtCore, uic, QtWidgets
from PyQt5.QtCore import QRunnable, QThreadPool, pyqtSignal, QObject
from pkg_resources import resource_filename
from functools import partial
import sys
import os.path as op
import os
import time
import requests
import logging
import traceback
import socket
import ulstools
import configdot

from .qt_dialogs import (
    OptionsDialog,
    qt_message_dialog,
    qt_yesno_dialog,
    ChooseSessionsDialog,
    qt_matplotlib_window,
    qt_dir_chooser,
)
from .qt_widgets import QtHandler, ProgressBar, ProgressSignals, XStream
from ulstools.num import check_hetu
from ..videos import _collect_session_videos, convert_videos
from .. import GaitDataError, nexus, cfg, sessionutils, envutils, c3d, stats, trial
from . import _tardieu
from ..autoprocess import (
    autoproc_session,
    autoproc_trial,
    automark_trial,
    _copy_session_videos,
)
from ..viz.plots import (
    plot_trials,
    plot_trial_timedep_velocities,
    plot_trial_velocities,
)
from ..viz.timedist import plot_session_average
from ..viz.plot_misc import _browse_localhost, _show_plotly_fig
from ..report import web, pdf

logger = logging.getLogger(__name__)

# setting this disables our internal handling of uncaught exceptions (so that the debugger
# can catch them) and also enables VSCode debugging of PyQt threads that we launch
# (ptvsd package is needed)
DEBUG_MODE = False
if DEBUG_MODE:
    import ptvsd


def _get_nexus_sessionpath():
    """Get Nexus sessionpath, handle exceptions for use outside _run_in_thread"""
    try:
        return nexus.get_sessionpath()
    except GaitDataError as e:
        _report_exception(e)
        return None


def _report_exception(e, title=None):
    """Report an exception via Qt dialog. Show title and exception message"""
    logger.debug('caught exception when running task')
    if title is None:
        title = 'There was an error running the operation. Details:'
    e_msg = _exception_msg(e)
    msg = '%s\n%s' % (title, e_msg)
    qt_message_dialog(msg)


def _exception_msg(e):
    """Return text representation of exception e"""
    # for our own error class, we know that a neat message is there
    # otherwise, we have no idea, so use generic repr()
    # XXX: some exception classes (e.g. IOError) repr doesn't print all relevant information
    # (e.g. the file name), however it looks like the safest choice (str() has its own issues)
    # this can probably be improved for Py3
    # for certain classes, str(e) works fine
    if (
        isinstance(e, TypeError)
        or isinstance(e, ValueError)
        or isinstance(e, GaitDataError)
    ):
        return str(e)
    else:
        return repr(e)


class PdfReportDialog(QtWidgets.QDialog):
    """Ask for patient/session info and report options"""

    def __init__(self, info, comparison=False, parent=None):
        super(self.__class__, self).__init__()
        self.comparison = comparison
        ui_filename = (
            'pdf_report_dialog_comparison.ui' if comparison else 'pdf_report_dialog.ui'
        )
        uifile = resource_filename('gaitutils', 'gui/%s' % ui_filename)
        uic.loadUi(uifile, self)
        # self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        if info is not None:
            if info['fullname'] is not None:
                self.lnFullName.setText(info['fullname'])
            if info['hetu'] is not None:
                self.lnHetu.setText(info['hetu'])
            if info['session_description'] is not None:
                self.lnDescription.setText(info['session_description'])

    def accept(self):
        """Update config and close dialog, if widget inputs are ok. Otherwise
        show an error dialog"""
        self.hetu = self.lnHetu.text()
        self.fullname = self.lnFullName.text()
        self.session_description = self.lnDescription.text()
        # take the page selections and write them into a dict (keys are named after the
        # selection widgets)
        self.pages = dict()
        for w in self.findChildren(QtWidgets.QWidget):
            wname = w.objectName()
            if wname[:2] == 'cb':
                self.pages[wname[2:]] = w.checkState()
        # require patient name if it's not a comparison report (which might compare different
        # patients); hetu needs to be valid if it's entered
        if (self.comparison or self.fullname) and (
            check_hetu(self.hetu) or not self.hetu
        ):
            self.done(QtWidgets.QDialog.Accepted)  # or call superclass accept
        else:
            qt_message_dialog(
                'Please enter a name. If hetu (Finnish SSN) is entered, it needs to be valid.'
            )


class WebReportInfoDialog(QtWidgets.QDialog):
    """Ask for patient info"""

    def __init__(self, info, parent=None, check_info=True):
        super(self.__class__, self).__init__()
        uifile = resource_filename('gaitutils', 'gui/web_report_info.ui')
        uic.loadUi(uifile, self)
        # self.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.check_info = check_info
        if info is not None:
            if info['fullname'] is not None:
                self.lnFullName.setText(info['fullname'])
            if info['hetu'] is not None:
                self.lnHetu.setText(info['hetu'])

    def accept(self):
        """Update config and close dialog, if widget inputs are ok. Otherwise
        show an error dialog"""
        self.recreate_plots = self.xbRecreatePlots.checkState()
        self.force_convert_videos = self.xbForceConvertVideos.checkState()
        self.video_only = self.xbVideoOnly.checkState()
        self.hetu = self.lnHetu.text().strip()
        self.fullname = self.lnFullName.text().strip()
        if self.check_info:
            ok = self.fullname and check_hetu(self.hetu)
        else:
            ok = not self.hetu or check_hetu(self.hetu)
        if ok:
            self.done(QtWidgets.QDialog.Accepted)  # or call superclass accept
        else:
            msg = 'Please enter a valid name and hetu'
            if not self.check_info:
                msg += ' (or leave empty)'
            qt_message_dialog(msg)


class WebReportDialog(QtWidgets.QDialog):
    """Dialog for managing web reports. In current implementation, this needs a
    GaitMenu instance as a parent (uses _run_in_thread() and other parent methods)"""

    def __init__(self, parent):
        super(self.__class__, self).__init__(parent)
        self.parent = parent
        # load user interface made with designer
        uifile = resource_filename('gaitutils', 'gui/web_report_dialog.ui')
        uic.loadUi(uifile, self)
        self.btnCreateReport.clicked.connect(lambda ev: self._create_web_report())
        self.btnDeleteReport.clicked.connect(self._delete_current_report)
        self.btnDeleteAllReports.clicked.connect(self._delete_all_reports)
        self.btnViewReport.clicked.connect(self._view_current_report)
        # add double click action to browse current report
        (
            self.listActiveReports.itemDoubleClicked.connect(
                lambda item: _browse_localhost(port=item.userdata)
            )
        )
        # these require active reports to be enabled
        self.reportWidgets = [
            self.btnDeleteReport,
            self.btnDeleteAllReports,
            self.btnViewReport,
        ]
        self._set_report_button_status()

    def _create_web_report(self, sessions=None):
        """Collect sessions, create the dash app, start it and launch a
        web browser on localhost on the correct port"""

        if self.listActiveReports.count() == cfg.web_report.max_reports:
            qt_message_dialog(
                'Maximum number of active web reports active. '
                'Please delete some reports first.'
            )
            return

        if sessions is None:
            dlg = ChooseSessionsDialog()
            if not dlg.exec_():
                return
            sessions = dlg.sessions
        report_name = web._report_name(sessions)
        existing_names = [item.text for item in self.listActiveReports.items]
        if report_name in existing_names:
            reply = qt_yesno_dialog(
                'There is already a report for %s. Recreate?' % report_name
            )
            if reply == QtWidgets.QMessageBox.YesRole:
                # delete the existing one
                theitem = [
                    item
                    for item in self.listActiveReports.items
                    if item.text == report_name
                ][0]
                self._delete_report(theitem)
            else:
                return

        # gather patient info files and merge them
        session_infos, info = sessionutils._merge_session_info(sessions)
        if info is None:
            qt_message_dialog(
                'Patient files do not match. Sessions may be '
                'from different patients. Continuing with default '
                'patient info.'
            )
            info = sessionutils.default_info()

        # if we could not load info, remember the previous values
        if info['fullname'] is None:
            info['fullname'] = self.parent._prev_info['fullname']
        if info['hetu'] is None:
            info['hetu'] = self.parent._prev_info['hetu']

        # ask user for info updates
        # dialog will be prepopulated with the values gathered above
        dlg_info = WebReportInfoDialog(info, check_info=False)
        if dlg_info.exec_():
            new_info = dict(
                hetu=dlg_info.hetu,
                fullname=dlg_info.fullname,
            )
            recreate_plots = dlg_info.recreate_plots
            force_convert_videos = dlg_info.force_convert_videos
            video_only = dlg_info.video_only
            info.update(new_info)

            # update info files in all sessions according to the user input
            # exclude the session specific keys
            for session in sessions:
                update_dict = dict(
                    fullname=dlg_info.fullname,
                    hetu=dlg_info.hetu,
                )
                session_infos[session].update(update_dict)
                sessionutils.save_info(session, session_infos[session])
            self.parent._prev_info.update(update_dict)
        else:
            return

        # give progress bar to parent so that parent can close it
        self.parent.prog = ProgressBar('Creating web report...')
        self.parent.prog.update('Collecting session information...', 0)
        signals = ProgressSignals()
        signals.progress.connect(lambda text, p: self.parent.prog.update(text, p))
        self.parent.prog._canceled.connect(signals.cancel)

        # for comparison reports, get representative trials only, unless doing
        # video only; this is to reduce the number of gait curves
        if len(sessions) > 1 and not video_only:
            tags = cfg.eclipse.repr_tags
        else:
            tags = cfg.eclipse.tags

        # collect all video files for conversion
        # includes tagged dynamic, video-only tagged, and static trials
        vidfiles = list()
        for session in sessions:
            vids = _collect_session_videos(session, tags=tags)
            vidfiles.extend(vids)

        if force_convert_videos or not convert_videos(vidfiles, check_only=True):
            self.parent._convert_vidfiles(vidfiles, signals)

        # launch the report creation thread
        self.parent._run_in_thread(
            web.dash_report,
            block_ui=True,
            finished_func=self.parent._enable_main_ui,
            result_func=self._web_report_ready,
            sessions=sessions,
            info=info,
            tags=tags,
            signals=signals,
            recreate_plots=recreate_plots,
            video_only=video_only,
        )

    @property
    def active_reports(self):
        """Return number of active web reports"""
        return self.listActiveReports.count()

    def shutdown(self):
        """Try to shutdown web servers"""
        # cannot use generator here since the loop changes the items
        for item in list(self.listActiveReports.items):
            self._delete_report(item)

    def _delete_report(self, item):
        """Shut down server for given list item, remove item"""
        port = item.userdata
        # compose url for shutdown request - see report.py
        url = 'http://127.0.0.1:%d/shutdown' % port
        # we have to make sure that localhost is not proxied
        proxies = {"http": None, "https": None}
        logger.debug('requesting server shutdown for port %d' % port)
        requests.get(url, proxies=proxies)
        self.listActiveReports.rm_current_item()

    def _delete_current_report(self):
        """Shut down server for current item, remove item"""
        item = self.listActiveReports.currentItem()
        if item is None:
            return
        msg = 'Are you sure you want to delete the report for %s?' % item.text
        reply = qt_yesno_dialog(msg)
        if reply == QtWidgets.QMessageBox.YesRole:
            self._delete_report(item)
        self._set_report_button_status()

    def _delete_all_reports(self):
        """Delete all web reports"""
        if self.listActiveReports.count() == 0:
            return
        msg = 'Are you sure you want to delete all reports?'
        reply = qt_yesno_dialog(msg)
        if reply != QtWidgets.QMessageBox.YesRole:
            return
        # cannot use generator here since the loop changes the items
        for item in list(self.listActiveReports.items):
            self._delete_report(item)
        self._set_report_button_status()

    def _view_current_report(self):
        """Open current report in browser"""
        item = self.listActiveReports.currentItem()
        if item is None:
            return
        port = item.userdata
        _browse_localhost(port=port)

    def _set_report_button_status(self):
        """Enable report buttons if reports exist, otherwise disable them"""
        n_reports = self.active_reports
        for widget in self.reportWidgets:
            widget.setEnabled(True if n_reports else False)

    def _web_report_ready(self, app):
        """Gets called when web report creation is successful. Open report in
        browser"""
        if app is None:
            # this should only happen when the report was cancelled, so we
            # exit quietly
            return
        # this code should automatically get a random port from the OS
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.bind(('', 0))
        s.listen(1)
        port = s.getsockname()[1]
        logging.debug('got port %d' % port)
        s.close()
        # The web servers need to go into separate threads/processes so that the
        # rest of the app can continue running. Due to difficulties of passing
        # data to processes, we use threads instead (the Qt threadpool). Each
        # running server then occupies a thread, this means that we need to
        # increase the threadpool max threads limit; otherwise new servers will
        # get queued by the threadpool and will not run.
        # Serving is a bit flaky in py2 (multiple requests cause exceptions)
        self.parent._run_in_thread(
            app.server.run, block_ui=False, debug=False, port=port, threaded=True
        )
        # double clicking on the list item will browse to corresponding port
        self.listActiveReports.add_item(app._gaitutils_report_name, data=port)
        # enable delete buttons etc.
        self._set_report_button_status()
        _browse_localhost(port=port)


class AddSessionDialog(QtWidgets.QDialog):
    """Dialog for adding trials to trials list"""

    def __init__(self, parent):
        QtWidgets.QDialog.__init__(self)
        uifile = resource_filename('gaitutils', 'gui/add_session_dialog.ui')
        uic.loadUi(uifile, self)

    def accept(self):
        self.c3ds = list()
        tags = cfg.eclipse.tags if self.rbAddTaggedTrials.isChecked() else None
        # get session
        if self.rbUseCurrentNexusSession.isChecked():
            session = _get_nexus_sessionpath()
        else:
            sessions = qt_dir_chooser()
            session = sessions[0] if sessions else None
        if session:
            self.c3ds = sessionutils.get_c3ds(
                session, tags=tags, trial_type='dynamic', check_if_exists=False
            )
            if self.c3ds:
                self.done(QtWidgets.QDialog.Accepted)
            else:
                qt_message_dialog('No trials found for session %s' % session)


class Gaitmenu(QtWidgets.QMainWindow):
    def __init__(self):
        super(self.__class__, self).__init__()
        # load user interface made with designer
        uifile = resource_filename('gaitutils', 'gui/gaitmenu.ui')
        uic.loadUi(uifile, self)

        # disable editing of log widget text
        self.txtOutput.setReadOnly(True)

        if (
            not cfg.general.allow_multiple_menu_instances
            and ulstools.env.already_running('gaitmenu')
        ):
            qt_message_dialog(
                'Another instance of the menu seems to be '
                'running. Please use that instance or '
                'stop it before starting a new one.'
            )
            sys.exit()

        if cfg.general.git_autoupdate:
            if envutils._git_autoupdate():
                # update was done, restart the script
                menu_script = op.join(envutils.pkg_dir, 'gui/gaitmenu.py')
                os.execv(sys.executable, ['python'] + [menu_script])

        self._web_report_dialog = WebReportDialog(self)

        # connect ui widgets
        self.btnAddSession.clicked.connect(self._add_session_dialog)
        self.btnAddTrials.clicked.connect(self._add_trials_dialog)
        self.btnAddNexusTrial.clicked.connect(self._add_nexus_trial)
        self.btnSelectAll.clicked.connect(self._select_all_trials)
        self.btnClearSelected.clicked.connect(self._remove_selected_trials)
        self.btnPlotTrials.clicked.connect(self._plot_selected_trials)
        self.btnAveragePlot.clicked.connect(self._average_trials)
        self.actionCreate_PDF_report.triggered.connect(
            lambda ev: self._create_pdf_report()
        )
        self.actionWeb_reports.triggered.connect(self._web_report_dialog.show)
        self.actionQuit.triggered.connect(self.close)
        self.actionOpts.triggered.connect(self._options_dialog)
        self.actionTardieu_analysis.triggered.connect(self._tardieu)
        self.actionExport_current_trial_to_EDF_format.triggered.connect(
            self._export_current_trial_to_edf
        )
        self.actionAutoprocess_session.triggered.connect(self._autoproc_session)
        self.actionAutoprocess_single_trial.triggered.connect(self._autoproc_trial)
        self.actionPDF_report_from_Nexus_session.triggered.connect(
            self._create_pdf_report_nexus
        )
        self.actionWeb_report_from_Nexus_session.triggered.connect(
            self._create_web_report_nexus
        )
        self.actionRun_postprocessing_pipelines.triggered.connect(
            self._postprocess_session
        )
        self.actionTrial_median_velocities.triggered.connect(
            self._plot_trial_median_velocities
        )
        self.actionTrial_timedep_velocities.triggered.connect(
            self._plot_trial_timedep_velocities
        )
        self.actionConvert_session_videos_to_web_format.triggered.connect(
            self._convert_session_videos
        )
        self.actionTime_distance_average.triggered.connect(self._plot_timedist_average)
        # XXX: these get run in the main thread
        self.actionCopy_session_videos_to_desktop.triggered.connect(
            _copy_session_videos
        )
        self.actionAutomark_events.triggered.connect(self._automark_trial)
        # trials table settings
        # force "item selected" style, otherwise it will depend on focus; set font size
        table_sheet = "QTableView{ selection-background-color: rgba(0, 0, 255, 50%); font-size: 8pt; }"
        self.tableTrials.setStyleSheet(table_sheet)
        self.tableTrials.cellDoubleClicked.connect(self._cell_doubleclicked)

        # set up radio buttons
        self.rb_map_backend = {'plotly': self.rbPlotly, 'matplotlib': self.rbMatplotlib}
        rb_backend_active = self.rb_map_backend[cfg.plot.backend]
        rb_backend_active.setChecked(True)

        # set up color/style comboboxes
        self.cbColorStyleVar.currentIndexChanged.connect(self._colorstylevar_changed)
        self.cbColorOrStyle.currentIndexChanged.connect(self._colorstylevar_changed)
        self.cbColorStyleBy.activated.connect(self._colorstyleby_changed)
        # dynamically create entries for vartype according to cfg
        for vartype in cfg.plot.color_by:
            vartype_txt = 'EMG' if vartype == 'emg' else vartype  # capitalize 'emg'
            self.cbColorStyleVar.addItem(vartype_txt)

        # add plot layouts to layout combobox
        cb_items = sorted(
            configdot.get_description(lo) or loname for loname, lo in cfg['layouts']
        )
        self.cbLayout.addItems(cb_items)
        # set default option to PiG lower body (if it's on the list)
        try:
            default_index = cb_items.index('PiG lower body kinematics')
        except ValueError:
            default_index = 0
        self.cbLayout.setCurrentIndex(default_index)
        # map descriptions to layout names
        self.layouts_map = {
            (configdot.get_description(lo) or loname): loname
            for loname, lo in cfg['layouts']
        }

        XStream.stdout().messageWritten.connect(self._log_message)
        XStream.stderr().messageWritten.connect(self._log_message)
        self.threadpool = QThreadPool()
        # we need a thread for each web server plus one worker thread
        self.threadpool.setMaxThreadCount(cfg.web_report.max_reports + 1)
        # keep refs to tardieu+mpl windows so they don't get garbage collected
        self._tardieuwin = None
        self._mpl_windows = list()
        # progress bar
        self.prog = None
        # used to save previous info dialog values
        self._prev_info = {'fullname': None, 'hetu': None}

    def _colorstylevar_changed(self, _idx):
        """Callback for color/style comboboxes.

        Updates the 'color/style by' combobox according to the new choice.
        """
        vartype, mode = self._get_colorstyle_choices()
        cfg_item = {'color': cfg.plot.color_by, 'style': cfg.plot.style_by}[mode]
        choice = cfg_item[vartype]
        if choice is None:
            choice = '(none)'
        idx = self.cbColorStyleBy.findText(choice)
        self.cbColorStyleBy.setCurrentIndex(idx)
        # disable styling for EMG variables
        enabled = not (vartype == 'emg' and mode == 'style')
        self.cbColorStyleBy.setEnabled(enabled)

    def _colorstyleby_changed(self, _idx):
        """Callback for color/style by xxx combobox.

        Changes the cfg item according to the current settings.
        """
        vartype, mode = self._get_colorstyle_choices()
        cfg_item = {'color': cfg.plot.color_by, 'style': cfg.plot.style_by}[mode]
        colorstyleby_choice = self.cbColorStyleBy.currentText()
        # Py2: convert unicode choices to str
        # (unicode prefix looks weird in Options interface)
        if sys.version_info.major == 2:
            colorstyleby_choice = str(colorstyleby_choice)
        if colorstyleby_choice == '(none)':
            colorstyleby_choice = None
        cfg_item[vartype] = colorstyleby_choice

    def _get_colorstyle_choices(self):
        """Get color/style choices from UI"""
        mode = self.cbColorOrStyle.currentText()
        vartype = self.cbColorStyleVar.currentText()
        vartype = 'emg' if vartype == 'EMG' else vartype  # undo capitalize 'emg'
        return vartype, mode

    @property
    def _plotting_backend(self):
        """Get backend selection from UI"""
        for backend, rb in self.rb_map_backend.items():
            if rb.isChecked():
                return backend

    def _automark_trial(self):
        session = _get_nexus_sessionpath()
        if session is None:
            return
        self._run_in_thread(automark_trial, finished_func=self._enable_main_ui)

    def _autoproc_session(self):
        """Wrapper to run autoprocess for Nexus session"""
        session = _get_nexus_sessionpath()
        if session is None:
            return
        c3ds = sessionutils.get_c3ds(session, trial_type='DYNAMIC')
        if c3ds:
            reply = qt_yesno_dialog(
                'Some of the dynamic trials have been '
                'processed already. Are you sure you want '
                'to run autoprocessing?'
            )
            if reply == QtWidgets.QMessageBox.NoRole:
                return

        self.prog = ProgressBar('Running autoprocessing...')
        signals = ProgressSignals()
        signals.progress.connect(lambda text, p: self.prog.update(text, p))
        self.prog._canceled.connect(signals.cancel)

        self._run_in_thread(
            autoproc_session, finished_func=self._enable_main_ui, signals=signals
        )

    def _autoproc_trial(self):
        """Wrapper to run autoprocess for Nexus trial"""

        self.prog = ProgressBar('Running autoprocessing...')
        signals = ProgressSignals()
        signals.progress.connect(lambda text, p: self.prog.update(text, p))
        self.prog._canceled.connect(signals.cancel)

        self._run_in_thread(
            autoproc_trial, finished_func=self._enable_main_ui, signals=signals
        )

    def _plot_timedist_average(self):
        """Plot time-distance average"""
        session = _get_nexus_sessionpath()
        if session is None:
            return
        self._run_in_thread(
            plot_session_average,
            finished_func=self._enable_main_ui,
            result_func=self._show_plots,
            session=session,
            backend=self._plotting_backend,
        )

    def _plot_trial_median_velocities(self):
        """Trial velocity plot from current Nexus session"""
        session = _get_nexus_sessionpath()
        if session is None:
            return
        self._run_in_thread(
            plot_trial_velocities,
            finished_func=self._enable_main_ui,
            result_func=self._show_plots,
            session=session,
            backend=self._plotting_backend,
        )

    def _plot_trial_timedep_velocities(self):
        """Trial velocity plot from current Nexus session"""
        session = _get_nexus_sessionpath()
        if session is None:
            return
        self._run_in_thread(
            plot_trial_timedep_velocities,
            finished_func=self._enable_main_ui,
            result_func=self._show_plots,
            session=session,
            backend=self._plotting_backend,
        )

    def _add_trial_to_table(self, tr):
        """Adds a trial to the trials table"""
        nrows = self.tableTrials.rowCount()
        self.tableTrials.insertRow(nrows)
        texts = (
            tr.trialname,
            tr.eclipse_data['DESCRIPTION'],
            tr.eclipse_data['NOTES'],
            tr.sessionpath,
        )
        for k, txt in enumerate(texts):
            item_ = QtWidgets.QTableWidgetItem(txt)
            if k == 0:
                # actual Trial instances are stored as userdata of
                # QTableWidgetItems on each rows 1st column; thus
                # we don't have to keep separate references to them
                item_.setData(QtCore.Qt.UserRole, tr)
            self.tableTrials.setItem(nrows, k, item_)

    def _add_c3dfiles(self, c3dfiles):
        """Add given c3d files to trials list"""
        c3dfiles = (op.normpath(fn) for fn in c3dfiles)
        self._disable_main_ui()  # in case it takes a while
        for c3dfile in c3dfiles:
            if not op.isfile(c3dfile):
                qt_message_dialog(
                    'Could not find the C3D file (%s) for this trial. Please make sure the trial has been processed and saved.'
                    % c3dfile
                )
                continue
            try:
                tr = trial.Trial(c3dfile)
            except GaitDataError as e:
                title = 'Could not load trial %s. Details:' % op.split(c3dfile)[-1]
                _report_exception(e, title=title)
            else:
                self._add_trial_to_table(tr)
                self.tableTrials.resizeColumnsToContents()
            finally:
                self._enable_main_ui()  # in case an unexpected exception is raised
        self._enable_main_ui()  # in case we do not hit the try/finally block

    def _add_nexus_trial(self):
        """Add directly from Nexus"""
        try:
            tr = trial.nexus_trial(from_c3d=cfg.trial.load_from_c3d)
        except GaitDataError as e:
            _report_exception(e)
        else:
            self._add_trial_to_table(tr)
            self.tableTrials.resizeColumnsToContents()

    def _add_session_dialog(self):
        """Show the add session dialog and add trials to list"""
        dlg = AddSessionDialog(self)
        if dlg.exec_():
            self._add_c3dfiles(dlg.c3ds)

    def _add_trials_dialog(self):
        """Add individual trials to list"""
        fout = QtWidgets.QFileDialog.getOpenFileNames(
            self, 'Load C3D files', None, 'C3D files (*.c3d)'
        )
        if fout[0]:
            self._add_c3dfiles(fout[0])

    def _select_all_trials(self):
        """Select all trials"""
        self.tableTrials.selectAll()

    @property
    def _selected_rows(self):
        """Return indices of selected rows"""
        return list(set(idx.row() for idx in self.tableTrials.selectedIndexes()))

    def _cell_doubleclicked(self, row, col):
        """Plot trial on double click"""
        tr = self.tableTrials.item(row, 0).data(QtCore.Qt.UserRole)
        self._plot_trials([tr])

    @property
    def _selected_trials(self):
        """Return list of trials that are currently selected"""
        sel_items = (self.tableTrials.item(row, 0) for row in self._selected_rows)
        return list(item.data(QtCore.Qt.UserRole) for item in sel_items)

    def _remove_selected_trials(self):
        """Remove selected trials from list"""
        while self._selected_rows:
            # this relies on _selected_rows dynamically changing after removals
            row = self._selected_rows[0]
            self.tableTrials.removeRow(row)

    def _export_current_trial_to_edf(self):
        """Export EMG from current trial to an EDF file"""
        if not self._selected_rows:
            qt_message_dialog('Load and select one or more trials first')
            return
        n_out = 0
        for tr in self._selected_trials:
            default_edfname = op.join(tr.sessionpath, tr.trialname + '.edf')
            if op.isfile(default_edfname):
                reply = qt_yesno_dialog(
                    'File %s already exists. Overwrite?' % default_edfname
                )
                if reply == QtWidgets.QMessageBox.NoRole:
                    continue
            tr.emg._edf_export(default_edfname)
            n_out += 1
        if n_out:
            qt_message_dialog(
                'Exported %d edf file(s) into %s' % (n_out, tr.sessionpath)
            )

    def _plot_selected_trials(self):
        if not self._selected_rows:
            return
        if any(tr.is_static for tr in self._selected_trials):
            qt_message_dialog(
                'One or more trials are static, plotting all trials as unnormalized'
            )
            self._plot_trials(self._selected_trials, normalized=False)
        else:
            self._plot_trials(self._selected_trials)

    def _plot_trials(self, trials, normalized=True):
        """Plot specified trials, or selected trials from menu"""
        have_avgtrials = any(isinstance(tr, stats.AvgTrial) for tr in trials)
        if not normalized or self.xbPlotUnnorm.checkState():
            if have_avgtrials:
                qt_message_dialog('Cannot plot average trials as unnormalized')
                return
            cycles = 'unnormalized'
        else:
            cycles = None
        layout_desc = self.cbLayout.currentText()
        layout_name = self.layouts_map[layout_desc]
        # FIXME: this is a bit kludgy
        is_emg_layout = 'EMG' in layout_name.upper()
        if self.xbEMGRMS.checkState():
            emg_mode = 'rms'
        elif have_avgtrials and is_emg_layout:
            qt_message_dialog('Averaged EMG can only be plotted in RMS mode')
            return
        else:
            emg_mode = None
        # FIXME: hardcoded legend type
        self._run_in_thread(
            plot_trials,
            finished_func=self._enable_main_ui,
            result_func=self._show_plots,
            trials=trials,
            layout_name=layout_name,
            backend=self._plotting_backend,
            cycles=cycles,
            emg_mode=emg_mode,
            legend_type='short_name_with_tag_and_cycle',
            auto_adjust_emg_layout=True,
        )

    def _average_trials(self):
        """Average trials from list, add resulting averaged trial to list"""
        if len(self._selected_rows) < 2:
            qt_message_dialog('Need at least 2 trials for averaging')
            return
        if any(isinstance(tr, stats.AvgTrial) for tr in self._selected_trials):
            qt_message_dialog('Cannot include averaged trials in average')
            return
        reject_outliers = cfg.trial.outlier_rejection_threshold
        self._run_in_thread(
            stats.AvgTrial.from_trials,
            finished_func=self._enable_main_ui,
            result_func=self._add_trial_to_table,
            trials=self._selected_trials,
            reject_outliers=reject_outliers,
        )

    def _create_web_report_nexus(self):
        """Create web report based on current Nexus session"""
        session = _get_nexus_sessionpath()
        if session:
            self._web_report_dialog._create_web_report(sessions=[session])

    def _show_plots(self, fig, backend=None):
        """Shows fig"""
        # use UI backend selection if unspecified
        if backend is None:
            backend = self._plotting_backend
        if backend == 'matplotlib':
            _mpl_win = qt_matplotlib_window(fig)
            self._mpl_windows.append(_mpl_win)
        elif backend == 'plotly':
            _show_plotly_fig(fig)

    def _convert_vidfiles(self, vidfiles, signals):
        """Convert given list of video files"""
        # get the converter processes
        procs = convert_videos(vidfiles=vidfiles)
        if not procs:
            logger.warning('video converter processes could not be started')
            return
        completed = False
        # wait in sleep loop until all converter processes have finished
        while not completed:
            if signals.canceled:
                logger.debug('canceled, killing video converter processes')
                for p in procs:
                    p.kill()
                break
            n_complete = len([p for p in procs if p.poll() is not None])
            prog_txt = 'Converting videos: %d of %d files done' % (
                n_complete,
                len(procs),
            )
            prog_p = 100 * n_complete / float(len(procs))
            signals.progress.emit(prog_txt, prog_p)
            time.sleep(0.1)
            completed = n_complete == len(procs)

    def _convert_session_videos(self):
        """Convert Nexus session videos to web format"""
        session = _get_nexus_sessionpath()
        if session is None:
            return
        try:
            vidfiles = _collect_session_videos(session, tags=cfg.eclipse.tags)
        except GaitDataError as e:
            qt_message_dialog(_exception_msg(e))
            return
        if not vidfiles:
            qt_message_dialog('Cannot find any video files for session %s' % session)
            return
        if convert_videos(vidfiles, check_only=True):
            reply = qt_yesno_dialog(
                'It looks like the session videos have already been converted. Redo?'
            )
            if reply == QtWidgets.QMessageBox.NoRole:
                return
        self._disable_main_ui()
        self.prog = ProgressBar('Converting session videos...')
        signals = ProgressSignals()
        signals.progress.connect(lambda text, p: self.prog.update(text, p))
        self.prog._canceled.connect(signals.cancel)
        self._convert_vidfiles(vidfiles, signals)
        self._enable_main_ui()

    def _postprocess_session(self):
        """Run additional postprocessing pipelines for tagged trials"""

        def _run_postprocessing():
            """Helper function that will be run in a separate thread"""
            nexus._close_trial()
            for k, tr in enumerate(trials, 1):
                trbase = op.splitext(tr)[0]
                vicon.OpenTrial(trbase, cfg.autoproc.nexus_timeout)
                nexus._run_pipelines_multiprocessing(cfg.autoproc.postproc_pipelines)
                prog_txt = 'Running postprocessing pipelines: %s for %d trials' % (
                    cfg.autoproc.postproc_pipelines,
                    len(trials),
                )
                prog_p = 100 * k / float(len(trials))
                signals.progress.emit(prog_txt, prog_p)
                if signals.canceled:
                    logger.debug('postprocessing pipelines were canceled')
                    return

        session = _get_nexus_sessionpath()
        if session is None:
            return
        # XXX: run for tagged + static - maybe this should be configurable
        trials = sessionutils.get_c3ds(
            session, tags=cfg.eclipse.tags, trial_type='dynamic', check_if_exists=False
        )
        trials += sessionutils.get_c3ds(
            session, trial_type='static', check_if_exists=False
        )
        if trials and cfg.autoproc.postproc_pipelines:
            logger.debug('running postprocessing for %s' % trials)
            vicon = nexus.viconnexus()
            self.prog = ProgressBar('Running postprocessing pipelines...')
            self.prog.update(
                'Running postprocessing pipelines: %s for %d '
                'trials' % (cfg.autoproc.postproc_pipelines, len(trials)),
                0,
            )
            signals = ProgressSignals()
            signals.progress.connect(lambda text, p: self.prog.update(text, p))
            self.prog._canceled.connect(signals.cancel)
            self._run_in_thread(
                _run_postprocessing, block_ui=True, finished_func=self._enable_main_ui
            )
        elif not trials:
            qt_message_dialog('No trials in session to run postprocessing for')
        elif not cfg.autoproc.postproc_pipelines:
            qt_message_dialog('No postprocessing pipelines defined')

    def closeEvent(self, event):
        """Confirm and close application."""

        if self._web_report_dialog.active_reports:
            reply = qt_yesno_dialog(
                'There are active web reports which '
                'will be closed. Are you sure you '
                'want to quit?'
            )
            if reply == QtWidgets.QMessageBox.YesRole:
                self._web_report_dialog.shutdown()
                self._close_mpl_windows()
                event.accept()
            else:
                event.ignore()
        else:
            self._close_mpl_windows()
            event.accept()

    def _close_mpl_windows(self):
        """Close all matplotlib windows"""
        for win in self._mpl_windows:
            win.close()

    def _options_dialog(self):
        """Show the options dialog"""
        dlg = OptionsDialog(self)
        dlg.exec_()

    def _create_pdf_report_nexus(self):
        """Create PDF report from Nexus session"""
        session = _get_nexus_sessionpath()
        if session is None:
            return
        self._create_pdf_report([session])

    def _create_pdf_report(self, sessions=None):
        """Create comparison or single session pdf report"""

        if sessions is None:
            dlg = ChooseSessionsDialog()
            if not dlg.exec_():
                return
            sessions = dlg.sessions
        comparison = len(sessions) > 1

        # gather patient info files and merge them
        session_infos, info = sessionutils._merge_session_info(sessions)
        if info is None:  # merge failed
            qt_message_dialog(
                'Patient files do not match. Sessions may be '
                'from different patients. Continuing with default '
                'patient info.'
            )
            info = sessionutils.default_info()

        # if we could not load info, remember the previous values
        if info['fullname'] is None:
            info['fullname'] = self._prev_info['fullname']
        if info['hetu'] is None:
            info['hetu'] = self._prev_info['hetu']

        if comparison:
            # compose a default description for comparison reports
            # we use the session_description field to pass this on to dialog and
            # report creation function
            descs = [_info['session_description'] for _info in session_infos.values()]
            if all(descs):  # use session descriptions if they exist
                info['session_description'] = ' vs. '.join(descs)
            else:  # we don't have description for every session - use dir names instead
                _session_dirs = (op.split(_session)[-1] for _session in sessions)
                info['session_description'] = ' vs. '.join(_session_dirs)

        # get inputs from user
        dlg_info = PdfReportDialog(info, comparison=comparison)
        if dlg_info.exec_():  # dialog was accepted
            new_info = dict(
                hetu=dlg_info.hetu,
                fullname=dlg_info.fullname,
                session_description=dlg_info.session_description,
            )
            info.update(new_info)

            # update info files in session dirs (except session specific keys)
            for session in sessions:
                update_dict = dict(fullname=dlg_info.fullname, hetu=dlg_info.hetu)
                # for single session reports, update also the session description
                if not comparison:
                    update_dict['session_description'] = info['session_description']
                session_infos[session].update(update_dict)
                sessionutils.save_info(session, session_infos[session])
                # update the remembered previous values
                self._prev_info.update(update_dict)
        else:
            return

        # create the report
        kwargs = {
            'info': info,
            'pages': dlg_info.pages,
            'finished_func': self._enable_main_ui,
            #'result_func': qt_message_dialog,  # show a message on success
        }
        if comparison:
            fun = pdf.create_comparison_report
            kwargs['sessionpaths'] = sessions
        else:
            fun = pdf.create_report
            kwargs['sessionpath'] = sessions[0]
            kwargs['write_timedist'] = True
            kwargs['write_extracted'] = True
        self._run_in_thread(fun, **kwargs)

    def _log_message(self, msg):
        """Log a message to the log widget"""
        c = self.txtOutput.textCursor()
        c.movePosition(QtGui.QTextCursor.End)
        self.txtOutput.setTextCursor(c)
        self.txtOutput.insertPlainText(msg)
        self.txtOutput.ensureCursorVisible()

    def _disable_main_ui(self):
        """Disable all operation buttons"""
        QtWidgets.QApplication.setOverrideCursor(QtCore.Qt.WaitCursor)
        self.setEnabled(False)  # disables whole main window
        # update display immediately in case thread gets blocked
        QtWidgets.QApplication.processEvents()

    def _enable_main_ui(self):
        """Enable all operation buttons, close progress bar if any and restore cursor."""
        if self.prog is not None:
            self.prog.reset()
        QtWidgets.QApplication.restoreOverrideCursor()
        self.setEnabled(True)

    def _tardieu(self):
        """Open the Tardieu window if it is not currently open"""
        if self._tardieuwin is None or not self._tardieuwin.isVisible():
            self._tardieuwin = _tardieu.TardieuWindow()
            self._tardieuwin.show()

    def _run_in_thread(
        self, fun, block_ui=True, finished_func=None, result_func=None, **kwargs
    ):
        """Run function fun with args kwargs in a worker thread.

        If block_ui==True, disable main ui until worker thread is finished.
        finished_func will be called when thread is finished. result_func
        will be called with the function return value as its single argument,
        unless an exception is raised during thread execution."""
        fun_ = partial(fun, **kwargs)
        if block_ui:
            self._disable_main_ui()
        self.runner = Runner(fun_)
        if finished_func:
            self.runner.signals.finished.connect(finished_func)
        if result_func:
            self.runner.signals.result.connect(lambda r: result_func(r))
        self.runner.signals.error.connect(lambda e: _report_exception(e))
        self.threadpool.start(self.runner)


class RunnerSignals(QObject):
    """Need a separate class since QRunnable cannot emit signals"""

    finished = pyqtSignal()  # thread finished
    result = pyqtSignal(object)  # successful completion - return value
    error = pyqtSignal(Exception)  # exception raised during run


class Runner(QRunnable):
    """Encapsulates threaded functions for QThreadPool"""

    def __init__(self, fun):
        super(Runner, self).__init__()
        self.fun = fun
        self.signals = RunnerSignals()

    def run(self):
        if DEBUG_MODE:  # do not handle exceptions, so that debugger can catch them
            # the ptvsd call is currently needed to debug 'native threads'
            # (apparently PyQt threads are such)
            ptvsd.debug_this_thread()
            retval = self.fun()
            self.signals.result.emit(retval)
            self.signals.finished.emit()
        else:  # the regular version - catch all exceptions and report via gui
            try:
                retval = self.fun()
            except Exception as e:
                self.signals.error.emit(e)
            else:
                self.signals.result.emit(retval)
            finally:
                self.signals.finished.emit()


def main():

    app = QtWidgets.QApplication(sys.argv)

    def my_excepthook(type_, value, tback):
        """Custom handler for unhandled exceptions:
        report to user via GUI and terminate."""
        tb_full = u''.join(traceback.format_exception(type_, value, tback))
        qt_message_dialog(
            'Oops! An unhandled exception was generated. '
            'The application will be closed.\n\n %s' % tb_full
        )
        # dump traceback to file
        # try:
        #    with io.open(Config.traceback_file, 'w', encoding='utf-8') as f:
        #        f.write(tb_full)
        # here is a danger of infinitely looping the exception hook,
        # so try to catch any exceptions...
        # except Exception:
        #    print('Cannot dump traceback!')
        sys.__excepthook__(type_, value, tback)
        app.quit()

    if not DEBUG_MODE:
        # for normal use, install our own handler for uncaught exceptions
        sys.excepthook = my_excepthook

    # add the Qt logging handler to root logger
    # it shows log messages in our QTextEdit widget
    root_logger = logging.getLogger()
    handler = QtHandler()
    handler.setFormatter(logging.Formatter("%(name)s: %(message)s"))
    root_logger.addHandler(handler)
    root_logger.setLevel(logging.DEBUG)

    # quiet down some noisy loggers
    logging.getLogger('matplotlib').setLevel(logging.WARNING)
    logging.getLogger('PyQt5.uic').setLevel(logging.WARNING)
    logging.getLogger('werkzeug').setLevel(logging.WARNING)

    gaitmenu = Gaitmenu()
    gaitmenu.show()
    logger.debug('Python interpreter: %s' % sys.executable)
    logger.debug('Python version: %s' % sys.version)
    logger.debug('Package directory: %s' % envutils.pkg_dir)
    if envutils.git_mode:
        logger.debug('Running from a git repository')
    else:
        logger.debug('Running from a pip install')
    if cfg.general.git_autoupdate:
        logger.debug('git autoupdate enabled')
    logger.debug('Nexus SDK path: %s' % nexus.nexus_path)
    if not c3d.BTK_IMPORTED:
        logger.warning('cannot find btk module; unable to read .c3d files')
    nexus_status = 'Vicon Nexus is %srunning' % ('' if nexus._nexus_pid() else 'not ')
    logger.debug(nexus_status)
    app.exec_()



# -*- coding: utf-8 -*-
"""
Session related functions


@author: Jussi (jnu@iki.fi)
"""

from builtins import str
import io
import os.path as op
import json
import datetime
import glob
import re
import logging

from .eclipse import get_eclipse_keys
from .config import cfg
from .envutils import GaitDataError


logger = logging.getLogger(__name__)

# the patient info keys
json_keys = ['fullname', 'hetu', 'session_description']
# exceptions that may occur when saving/loading json
json_exceptions = (UnicodeDecodeError, EOFError, IOError, TypeError, ValueError)


def load_quirks(session):
    """Load quirks JSON file.

    Quirks represent session-specific oddities that have to be taken into account
    when loading trials, e.g. wrong EMG scaling etc.

    Currently supported quirks are:

    emg_correction_factor : float
        All EMG data will be multiplied by this factor after reading.
    ignore_eclipse_fp_info : bool
        If True, Eclipse forceplate fields will be ignored when loading trials from
        the session.

    Parameters
    ----------
    session : str
        The session path.

    Returns
    -------
    dict
        The quirks in a dict.
    """

    quirks = dict()
    fname = op.join(session, 'quirks.json')
    if op.isfile(fname):
        with io.open(fname, 'r', encoding='utf-8') as f:
            try:
                quirks.update(json.load(f))
            except json_exceptions:
                logger.warning('cannot load quirks file %s' % fname)
    return quirks


def default_info():
    """Return a default patient info dict.

    Returns
    -------
    dict
        The info dict with all keys set to None.
    """
    return {key: None for key in json_keys}


def load_info(session):
    """Load the patient info dict from a given session

    Parameters
    ----------
    session : str
        The session path.

    Returns
    -------
    dict
        The patient info.
    """
    fname = op.join(session, 'patient_info.json')
    if op.isfile(fname):
        with io.open(fname, 'r', encoding='utf-8') as f:
            try:
                info = json.load(f)
                # check for extra keys (e.g. obsoleted ones), do not return them
                extra_keys = set(info.keys()) - set(json_keys)
                if extra_keys:
                    logger.warning(
                        'Extra keys %s in patient info file %s' % (extra_keys, fname)
                    )
                    for key in extra_keys:
                        info.pop(key)
                missing_keys = set(json_keys) - set(info.keys())
                if missing_keys:
                    logger.warning(
                        'Missing keys %s in patient info file %s'
                        % (missing_keys, fname)
                    )
                    # supply default values for missing keys
                    for key in missing_keys:
                        info[key] = default_info()[key]
            except json_exceptions:
                raise GaitDataError('Error loading patient info file %s' % fname)
    else:
        info = None
    return info


def save_info(session, patient_info):
    """Save an info dict into a session.

    Parameters
    ----------
    session : str
        The session path.
    patient_info : dict
        The patient info.
    """
    fname = op.join(session, 'patient_info.json')
    try:
        with io.open(fname, 'w', encoding='utf-8') as f:
            f.write(str(json.dumps(patient_info, ensure_ascii=False)))
    except json_exceptions:
        raise GaitDataError('Error saving patient info file %s ' % fname)


def _merge_session_info(sessions):
    """Merge patient info files across sessions.
    The fullname and hetu keys must match.
    Returns a 2-tuple with: dict of individual session infos, merged info"""
    session_infos = {
        session: (load_info(session) or default_info()) for session in sessions
    }
    info = default_info()
    # merging session descriptions does not really make sense, but do it for
    # consistency (so we can also call this for single session)
    for key in json_keys:
        allvals = set([session_infos[session][key] for session in sessions])
        if None in allvals:
            allvals.remove(None)
        if key == 'fullname' or key == 'hetu':
            if len(allvals) > 1:
                logger.warning('name / hetu do not match across sessions')
                return session_infos, None
        # in case of conflicts, this just picks from the last session
        info[key] = allvals.pop() if allvals else None
    return session_infos, info


def enf_to_trialfile(fname, ext):
    """Convert the name of a trial .enf file to another type of trial file.

    Parameters
    ----------
    fname : str
        The .enf file name.
    ext : str
        File extension, e.g. 'c3d'. Can be supplied with a leading dot.

    Returns
    -------
    str
        The converted filename.
    """
    if ext[0] == '.':
        ext = ext[1:]
    enfre = r'\.*.Trial\d*.enf'  # .Trial followed by zero or more digits
    res = re.search(enfre, fname)
    if res is None:
        raise GaitDataError('Filename %s is not a trial .enf' % fname)
    return fname.replace(res.group(), '.%s' % ext)


def get_session_date(sessionpath):
    """Get the date when the session was recorded (datetime.datetime object).

    Parameters
    ----------
    sessionpath : str
        The session path.

    Returns
    -------
    datetime.datetime
        The datetime.
    """
    enfs = get_enfs(sessionpath)
    x1ds = [enf_to_trialfile(fn, 'x1d') for fn in enfs]
    if not x1ds:
        raise GaitDataError('Invalid session %s' % sessionpath)
    else:
        x1d = x1ds[0]
        if not op.isfile(x1d):
            raise GaitDataError('x1d trial file %s does not exist' % x1d)
        return datetime.datetime.fromtimestamp(op.getmtime(x1d))


def _get_session_enfs(sessionpath):
    """Return list of .enf files for the session """
    enfglob = op.join(sessionpath, '*Trial*.enf')
    for enf in glob.iglob(enfglob):
        yield enf


def _filter_by_eclipse_keys(enfs, patterns, eclipse_keys):
    """Filter for enfs whose Eclipse key values match given patterns."""
    if not isinstance(patterns, list):
        patterns = [patterns]
    if not isinstance(eclipse_keys, list):
        eclipse_keys = [eclipse_keys]
    for enf in enfs:
        ecldi = {key.upper(): val.upper() for key, val in get_eclipse_keys(enf).items()}
        eclvals = [val for key, val in ecldi.items() if key in eclipse_keys]
        for pattern in patterns:
            if any(pattern.upper() in eclval for eclval in eclvals):
                yield enf


def _filter_by_tags(enfs, tags):
    """Filter by given tags"""
    return _filter_by_eclipse_keys(enfs, tags, cfg.eclipse.tag_keys)


def _filter_by_type(enfs, trial_type):
    """Filter by trial type"""
    return _filter_by_eclipse_keys(enfs, trial_type, 'TYPE')


def _filter_to_c3ds(enfs):
    """Convert enf filenames to c3d"""
    for enf in enfs:
        yield enf_to_trialfile(enf, 'c3d')


def _filter_exists(files):
    for f in files:
        if op.isfile(f):
            yield f


def get_enfs(sessionpath, tags=None, trial_type=None, check_if_exists=True):
    """Get specified enf files for a session.

    Parameters
    ----------
    sessionpath : str
        The session path.
    tags : list, optional
        List of Eclipse tags to filter for. E.g. ['T1'] would return only .enf
        files that are tagged with 'T1' in Eclipse. An empty list or None
        disables filtering by tags.
    trial_type : str, optional
        Trial type (Eclipse TYPE field), e.g. 'static' or 'dynamic'.
    check_if_exists : bool, optional
        If True, return only enf files that actually exist.

    Returns
    -------
    list
        List of enf files.
    """

    enfs = _get_session_enfs(sessionpath)
    if trial_type is not None:
        enfs = _filter_by_type(enfs, trial_type)
    if tags:
        enfs = _filter_by_tags(enfs, tags)
    if check_if_exists:
        enfs = _filter_exists(enfs)
    return list(enfs)


def get_c3ds(sessionpath, tags=None, trial_type=None, check_if_exists=True):
    """Get c3d files for a session. Similar to get_enfs above.

    Parameters
    ----------
    sessionpath : str
        The session path.
    tags : list, optional
        List of Eclipse tags to filter for. E.g. ['T1'] would return only .enf
        files that are tagged with 'T1' in Eclipse. An empty list or None
        disables filtering by tags.
    trial_type : str, optional
        Trial type (Eclipse TYPE field), e.g. 'static' or 'dynamic'.
    check_if_exists : bool, optional
        If True, return only c3d files that actually exist.

    Returns
    -------
    list
        List of c3d files.
    """

    enfs = get_enfs(
        sessionpath, tags=tags, trial_type=trial_type, check_if_exists=check_if_exists
    )
    c3ds = _filter_to_c3ds(enfs)
    if check_if_exists:
        c3ds = _filter_exists(c3ds)
    return list(c3ds)


def _get_tagged_dynamic_c3ds_from_sessions(sessions, tags=None):
    """Gather all tagged dynamic c3d files from given sessions."""
    if not isinstance(sessions, list):
        sessions = [sessions]
    c3ds_all = list()
    for session in sessions:
        c3ds = get_c3ds(session, tags=tags, trial_type='dynamic')
        # we require at least on tagged trial for each session
        if not c3ds:
            raise GaitDataError('No tagged trials found for session %s' % session)
        c3ds_all.extend(c3ds)
    return c3ds_all



# -*- coding: utf-8 -*-
"""

misc plotting related stuff

@author: Jussi (jnu@iki.fi)
"""

import plotly
from matplotlib.figure import Figure
from PyQt5 import QtWidgets
import subprocess
import logging

from . import plot_matplotlib, plot_plotly
from ..config import cfg
from ..gui import qt_dialogs
from ..envutils import _named_tempfile

logger = logging.getLogger(__name__)


def _show_plotly_fig(fig):
    """Show a Plotly fig in the configured browser"""
    tmp_html = _named_tempfile('.html')
    plotly.offline.plot(fig, filename=tmp_html, auto_open=False, validate=False)
    _browse_localhost(url='file:///%s' % tmp_html)


def _browse_localhost(url=None, port=None):
    """Open configured browser on url or localhost:port"""
    if not url:
        if port:
            url = '127.0.0.1:%d' % port
        else:
            raise ValueError('neither url nor valid localhost port specified')
    try:
        proc = subprocess.Popen([cfg.general.browser_path, url])
        logger.debug('new browser pid %d' % proc.pid)
    except Exception:
        raise RuntimeError(
            'Cannot start configured web browser: %s' % cfg.general.browser_path
        )


def get_backend(backend_name):
    """Return plotting backend module.

    Parameters
    ----------
    backend_name : str | None
        Name of the backend. Currently supported backends are 'plotly' and
        'matplotlib'. If None, return configured default backend.

    Returns
    -------
    backend : module
        The module.
    """
    if backend_name is None:
        backend_name = cfg.plot.backend
    backends = {'plotly': plot_plotly, 'matplotlib': plot_matplotlib}
    if backend_name not in backends:
        raise ValueError('no such plotting backend %s' % backend_name)
    return backends[backend_name]


def show_fig(fig):
    """Show a figure created by the plotting functions.

    What actually happens depends on the backend. Plotly figures are shown in a
    browser window. For matplotlib plots, a new Qt window is created and the
    plot is shown there.

    Parameters
    ----------
    fig : Figure | dict
        The figure to show.
    """
    if not fig:
        raise ValueError('No figure to show')
    if isinstance(fig, Figure):  # matplotlib
        app = QtWidgets.QApplication([])
        win = qt_dialogs.qt_matplotlib_window(fig)
        app.exec_()
    else:  # plotly
        # plotly figures may be of different types (list of traces etc.)
        # so just lazily hope that it's actually a plotly figure
        _show_plotly_fig(fig)



# -*- coding: utf-8 -*-
"""
Handles gaitutils config files.

@author: Jussi (jnu@iki.fi)
"""
from __future__ import print_function

from builtins import str
import os.path as op
import io
import logging
from pkg_resources import resource_filename
from configdot import parse_config, update_config, dump_config


# logging handlers might not be installed at this point, so config related
# messages may not be seen at all. alternative would be to use print
logger = logging.getLogger(__name__)


def _handle_cfg_defaults(cfg):
    """Handle deprecated and default config values"""
    if not isinstance(cfg.plot.emg_yscale, float):
        ysc = cfg.plot.emg_yscale[1]
        logger.warning('emg_yscale was changed to float, using %g' % ysc)
        cfg.plot.emg_yscale = str(cfg.plot.emg_yscale[1])
    if cfg.general.normaldata_files == 'default':
        fn = resource_filename('gaitutils', 'data/normal.gcd')
        cfg.general.normaldata_files = [fn]
    if cfg.emg.normaldata_file == 'default':
        fn = resource_filename('gaitutils', 'data/emg_normaldata.json')
        cfg.emg.normaldata_file = fn
    if cfg.general.videoconv_path == 'default':
        fn = resource_filename('gaitutils', 'thirdparty/ffmpeg2theora.exe')
        cfg.general.videoconv_path = fn
    if cfg.autoproc.write_eclipse_fp_info is True:
        cfg.autoproc.write_eclipse_fp_info = 'write'


# location of the default config file
cfg_template_fn = resource_filename(__name__, 'data/default.cfg')
# Location of the user specific config file. On Windows, this typically puts the
# config at C:\Users\Username, since the USERPROFILE environment variable points
# there. Putting the config in a networked home dir requires some tinkering with
# environment variables (e.g. setting HOME)
homedir = op.expanduser('~')
cfg_user_fn = op.join(homedir, '.gaitutils.cfg')

# provide the global cfg instance
# read template config
cfg = parse_config(cfg_template_fn)
if op.isfile(cfg_user_fn):
    logger.debug('reading user config from %s' % cfg_user_fn)
    cfg_user = parse_config(cfg_user_fn)
    # update config from user file, but do not overwrite comments
    # new config items are only allowed in layouts section
    update_config(
        cfg,
        cfg_user,
        create_new_sections=False,
        create_new_items=['layouts'],
        update_comments=False,
    )
else:
    logger.warning('no config file, trying to create %s' % cfg_user_fn)
    cfg_txt = dump_config(cfg)
    with io.open(cfg_user_fn, 'w', encoding='utf8') as f:
        f.writelines(cfg_txt)

_handle_cfg_defaults(cfg)

# if using print for config messages:
# sys.stdout.flush()  # make sure that warnings are printed out



# -*- coding: utf-8 -*-
"""
Compute statistics across/within trials


@author: Jussi (jnu@iki.fi)
"""

from __future__ import division

import logging
import numpy as np
import scipy
import os.path as op
import itertools
from collections import defaultdict

from .trial import Trial, Gaitcycle
from . import models, numutils
from .envutils import GaitDataError
from .numutils import _get_local_max, _get_local_min
from .config import cfg
from .emg import AvgEMG


logger = logging.getLogger(__name__)


class AvgTrial(Trial):
    """Gait trial -style class that holds averaged data. The API is designed to mimic trial.Trial.

    An AvgTrial instance can be created from averaged data using __init__()
    or from a list of trials using from_trials().
    """

    def __repr__(self):
        s = '<AvgTrial |'
        s += ' trial name: %s' % self.trialname
        s += ', trials: %s' % self.source
        s += '>'
        return s

    @classmethod
    def from_trials(
        cls,
        trials,
        sessionpath=None,
        reject_zeros=None,
        reject_outliers=None,
        use_medians=None,
    ):
        """Build AvgTrial from a list of trials"""
        nfiles = len(trials)
        data_all, cycles = collect_trial_data(trials)

        avgdata_model, stddata_model, ncycles_ok_analog = average_model_data(
            data_all['model'],
            reject_zeros=reject_zeros,
            reject_outliers=reject_outliers,
            use_medians=use_medians,
        )
        avgdata_emg, stddata_emg, ncycles_ok_emg = average_analog_data(
            data_all['emg'],
            rms=True,
            reject_outliers=reject_outliers,
            use_medians=use_medians,
        )

        return cls(
            avgdata_model=avgdata_model,
            stddata_model=stddata_model,
            avgdata_emg=avgdata_emg,
            stddata_emg=stddata_emg,
            sessionpath=sessionpath,
            nfiles=nfiles,
        )

    def __init__(
        self,
        avgdata_model=None,
        stddata_model=None,
        avgdata_emg=None,
        stddata_emg=None,
        sessionpath=None,
        nfiles=None,
    ):
        if avgdata_model is None and avgdata_emg is None:
            raise ValueError('no data for average')

        if nfiles is None:
            raise ValueError('nfiles must be supplied')

        self.nfiles = nfiles
        if sessionpath:
            self.sessionpath = sessionpath
            self.sessiondir = op.split(sessionpath)[-1]
            self.trialname = '%s avg. (%d trials)' % (self.sessiondir, self.nfiles)
        else:
            self.trialname = '%d trial avg.' % self.nfiles
            self.sessiondir = None
            self.sessionpath = None

        if avgdata_model is None:
            self._model_data = dict()
        else:
            self._model_data = avgdata_model
        self.stddev_data = stddata_model

        if avgdata_emg is None:
            avgdata_emg = dict()
        self.emg = AvgEMG(avgdata_emg)
        analog_npts = 1000  # default
        for ch in list(avgdata_emg.keys()):
            if avgdata_emg[ch] is not None:
                analog_npts = len(avgdata_emg[ch])

        self.tn_analog = np.linspace(0, 100, analog_npts)

        self.source = 'averaged data'
        self.name = 'Unknown'
        self.is_static = False
        # self.n_ok = n_ok  XXX
        self.t = np.arange(101)  # data is on normalized cycle 0..100%
        # fake 2 gait cycles, L/R
        self.cycles = list()
        self.cycles.append(
            Gaitcycle(0, 101, 60, 'R', True, 1, 1000, name='right', trial=self)
        )
        self.cycles.append(
            Gaitcycle(0, 101, 60, 'L', True, 1, 1000, name='left', trial=self)
        )
        self.ncycles = 2
        self.eclipse_data = defaultdict(lambda: '', {})
        self._marker_data = dict()

    def get_model_data(self, var, cycle=None):
        """Get averaged model variable.

        Parameters
        ----------
        var : str
            The variable name.
        """
        try:
            data = self._model_data[var]
        except KeyError:
            logger.info('no averaged data for %s, returning nans' % var)
            var_dims = (3, 101)  # is this universal?
            data = np.empty(var_dims)
            data[:] = np.nan
        return self.t, data

    def get_emg_data(self, ch, rms=None, cycle=None):
        """Get averaged EMG RMS data.

        Parameters
        ----------
        ch : str
            The channel name.
        """
        if not rms:
            raise ValueError('AvgTrial only supports EMG in RMS mode')
        return self.tn_analog, self.emg.get_channel_data(ch, rms=True)

    def get_marker_data(self, marker, cycle=None):
        raise GaitDataError('AvgTrial does not average marker data yet')


def _robust_reject_rows(data, p_threshold):
    """Reject rows (observations) from data based on robust Z-score"""
    # a Bonferroni type correction for the p-threshold
    p_threshold_corr = p_threshold / data.shape[1]
    # when computing outliers, estimate median absolute deviation
    # using all data, instead of a frame-based MAD estimates.
    # this is necessary since frame-based MAD values
    # can become really small especially in small datasets, causing
    # the Z-score to blow up and data getting rejected unnecessarily.
    outlier_inds = numutils.outliers(
        data, axis=0, single_mad=False, p_threshold=p_threshold_corr
    )
    outlier_rows = np.unique(outlier_inds[0])
    if outlier_rows.size > 0:
        logger.info(
            'rejected %d outlier(s) (corrected P=%g)'
            % (outlier_rows.size, p_threshold_corr)
        )
        data = np.delete(data, outlier_rows, axis=0)
    return data


def average_analog_data(data, rms=None, reject_outliers=None, use_medians=None):
    """Average collected analog data.

    Parameters
    ----------
    data : dict
        Data to average (from collect_trial_data)
    rms : bool
        Compute RMS before averaging.
    reject_outliers : float or None
        None for no automatic outlier rejection. Otherwise, a P value for false
        rejection (assuming strictly normally distributed data). Outliers are
        rejected based on robust statistics (modified Z-score).
    use_medians: bool
        Use median and MAD (median absolute deviation) instead of mean and
        stddev. The median is robust to outliers but not robust to small
        sample size. Thus, use of medians may be a bad idea for small samples.

    Returns
    -------
    avgdata : dict
        Averaged (or median) data as numpy array for each variable.
    stddata : dict
        Standard dev (or MAD) as numpy array for each variable.
    ncycles_ok : dict
        N of accepted cycles for each variable.
    """
    stddata = dict()
    avgdata = dict()
    ncycles_ok = dict()

    if rms is None:
        rms = False

    if use_medians is None:
        use_medians = False

    for var, vardata in data.items():
        if vardata is None:
            stddata[var] = None
            avgdata[var] = None
            ncycles_ok[var] = 0
            continue
        else:
            if rms:
                vardata = numutils.rms(vardata, cfg.emg.rms_win, axis=1)
            n_ok = vardata.shape[0]
            # do the outlier rejection
            if n_ok > 0 and reject_outliers is not None and not use_medians:
                rms_status = 'RMS for ' if rms else ''
                logger.info('averaging %s%s (N=%d)' % (rms_status, var, n_ok))
                vardata = _robust_reject_rows(vardata, reject_outliers)
            n_ok = vardata.shape[0]
            if n_ok == 0:
                stddata[var] = None
                avgdata[var] = None
            elif use_medians:
                stddata[var] = numutils.mad(vardata, axis=0)
                avgdata[var] = np.median(vardata, axis=0)
            else:
                stddata[var] = vardata.std(axis=0) if n_ok > 0 else None
                avgdata[var] = vardata.mean(axis=0) if n_ok > 0 else None
            ncycles_ok[var] = n_ok

    if not avgdata:
        logger.warning('nothing averaged')
    return (avgdata, stddata, ncycles_ok)


def average_model_data(data, reject_zeros=None, reject_outliers=None, use_medians=None):
    """Average collected model data.

    Parameters
    ----------
    data : dict
        Data to average (from collect_trial_data).
    reject_zeros : bool
        Reject any curves which contain zero values. Exact zero values are
        commonly used to mark gaps. No zero rejection is done for kinetic vars,
        since these may become zero due to clamping of force data.
    reject_outliers : float or None
        None for no automatic outlier rejection. Otherwise, a P value for false
        rejection (assuming strictly normally distributed data). Outliers are
        rejected based on robust statistics (modified Z-score).
    use_medians: bool
        Use median and MAD (median absolute deviation) instead of mean and
        stddev. The median is robust to outliers but not robust to small sample
        size. Thus, use of medians may be a bad idea for small samples.

    Returns
    -------
    avgdata : dict
        Averaged (or median) data as numpy array for each variable.
    stddata : dict
        Standard dev (or MAD) as numpy array for each variable.
    ncycles_ok : dict
        N of accepted cycles for each variable.
    """
    stddata = dict()
    avgdata = dict()
    ncycles_ok = dict()

    if use_medians is None:
        use_medians = False

    if reject_zeros is None:
        reject_zeros = True

    for var, vardata in data.items():
        if vardata is None:
            stddata[var] = None
            avgdata[var] = None
            ncycles_ok[var] = 0
            continue
        else:
            Ntot = vardata.shape[0]
            if reject_zeros:
                this_model = models.model_from_var(var)
                if not this_model.is_kinetic_var(var):
                    rows_bad = np.where(np.any(vardata == 0, axis=1))[0]
                    if len(rows_bad) > 0:
                        logger.info(
                            '%s: rejecting %d curves with zero values'
                            % (var, len(rows_bad))
                        )
                        vardata = np.delete(vardata, rows_bad, axis=0)
            n_ok = vardata.shape[0]
            # do the outlier rejection
            if n_ok > 0 and reject_outliers is not None and not use_medians:
                logger.info('%s:' % var)
                vardata = _robust_reject_rows(vardata, reject_outliers)
            n_ok = vardata.shape[0]
            if n_ok == 0:
                stddata[var] = None
                avgdata[var] = None
            elif use_medians:
                stddata[var] = numutils.mad(vardata, axis=0)
                avgdata[var] = np.median(vardata, axis=0)
            else:
                stddata[var] = vardata.std(axis=0) if n_ok > 0 else None
                avgdata[var] = vardata.mean(axis=0) if n_ok > 0 else None
            logger.debug('%s: averaged %d/%d curves' % (var, n_ok, Ntot))
            ncycles_ok[var] = n_ok

    if not avgdata:
        logger.warning('nothing averaged')
    return (avgdata, stddata, ncycles_ok)


def collect_trial_data(
    trials, collect_types=None, fp_cycles_only=None, analog_len=None
):
    """Read model and analog data across trials into numpy arrays.

    Parameters
    ----------
    trials : list | str | Trial
        List of c3d filenames or Trial instances to collect data from.
        Alternatively, a single filename or Trial instance.
    collect_types : list | None
        The types of data to collect. Currently supported types: 'model', 'emg'.
        If None, collect all supported types.
    fp_cycles_only : bool
        If True, collect data from forceplate cycles only. Kinetics model vars
        will always be collected from forceplate cycles only.
    analog_len : int
        Analog data length varies by gait cycle, so it will be resampled into
        grid length specified by analog_len (default 1000 samples)

    Returns
    -------
    tuple
        Tuple of (data_all, cycles_all):

            data_all : dict
                Nested dict of the collected data. First key is the variable type,
                second key is the variable name. The values are NxT ndarrays of the data,
                where N is the number of collected curves and T is the dimensionality.
            cycles_all : dict
                Nested dict of the collected cycles. First key is the variable type,
                second key is the variable name. The values are Gaitcycle instances.

        Example: you can obtain all collected curves for LKneeAnglesX as
        data_all['model']['LKneeAnglesX']. This will be a Nx101 ndarray. You can obtain
        the corresponding gait cycles as cycles_all['model']['LKneeAnglesX']. This will
        be a length N list of Gaitcycles. You can use that to obtain various metadata, e.g.
        create a list of toeoff frames for each curve:
        [cyc.toeoffn for cyc in cycles_all['model']['LKneeAnglesX']]
    """

    data_all = dict()
    cycles_all = dict()

    if fp_cycles_only is None:
        fp_cycles_only = False

    if collect_types is None:
        collect_types = ['model', 'emg']

    if analog_len is None:
        analog_len = 1000  # reasonable default for analog data (?)

    if not trials:
        return None, None

    if not isinstance(trials, list):
        trials = [trials]

    if 'model' in collect_types:
        data_all['model'] = defaultdict(lambda: None)
        cycles_all['model'] = defaultdict(list)
        models_to_collect = models.models_all
    else:
        models_to_collect = list()

    if 'emg' in collect_types:
        data_all['emg'] = defaultdict(lambda: None)
        cycles_all['emg'] = defaultdict(list)
        emg_chs_to_collect = cfg.emg.channel_labels.keys()
    else:
        emg_chs_to_collect = list()

    trial_types = list()
    for trial_ in trials:
        # create Trial instance in case we got filenames as args
        trial = trial_ if isinstance(trial_, Trial) else Trial(trial_)
        logger.info('collecting data for %s' % trial.trialname)
        trial_types.append(trial.is_static)
        if any(trial_types) and not all(trial_types):
            raise GaitDataError('Cannot mix dynamic and static trials')

        cycles = [None] if trial.is_static else trial.cycles

        for cycle in cycles:
            # collect model data
            for model in models_to_collect:
                for var in model.varnames:
                    if not trial.is_static:
                        # pick data only if var context matches cycle context
                        # FIXME: should implement context() for models
                        # (and a filter for context?)
                        if var[0] != cycle.context:
                            continue
                        # don't collect kinetics if cycle is not on forceplate
                        if (
                            model.is_kinetic_var(var) or fp_cycles_only
                        ) and not cycle.on_forceplate:
                            continue
                    _, data = trial.get_model_data(var, cycle)
                    if np.all(np.isnan(data)):
                        logger.debug('no data for %s/%s' % (trial.trialname, var))
                    else:
                        cycles_all['model'][var].append(cycle)
                        if var not in data_all['model']:
                            data_all['model'][var] = data[None, :]
                        else:
                            data_all['model'][var] = np.concatenate(
                                [data_all['model'][var], data[None, :]]
                            )

            for ch in emg_chs_to_collect:
                # check whether cycle matches channel context
                if not trial.is_static and not trial.emg.context_ok(ch, cycle.context):
                    continue
                # get data on analog sampling grid and compute rms
                try:
                    _, data = trial.get_emg_data(ch)
                except (KeyError, GaitDataError):
                    logger.warning('no channel %s for %s' % (ch, trial))
                    continue
                # resample to requested grid
                data_cyc = scipy.signal.resample(data, analog_len)
                cycles_all['emg'][ch].append(cycle)
                if ch not in data_all['emg']:
                    data_all['emg'][ch] = data_cyc[None, :]
                else:
                    data_all['emg'][ch] = np.concatenate(
                        [data_all['emg'][ch], data_cyc[None, :]]
                    )
    logger.info('collected %d trials' % len(trials))
    return data_all, cycles_all


def curve_extract_values(curves, toeoffs):
    """Extract values from gait curves.

    This extracts values such as swing phase maximum from a set of gait curves.
    The curves are input as ndarrays, returned by e.g. collect_trial_data().
    Data with any nans will result in (at least partially) nan output values.

    Parameters
    ----------
    curves : ndarray
        NxT array of gait curves. Typically T==101 for normalized data.
    toeoffs : ndarray
        Nx1 array of toeoff frame indices, one for each curve. This frame
        separates the contact phase from the swing phase.

    Returns
    -------
    dict
        Dictionary of results, with following keys:
            'contact' : list of curve values at initial foot contact (frame 0)
            'toeoff' : list of curve values at toeoff
            'extrema' : nested dict of simple extrema
            'peaks' : nested dict of peaks (local extrema)
        The nested dicts have keys:
            1: 'overall', 'swing', or 'stance' : the phase of the gait curve
            2: 'min', 'argmin', 'max', or 'argmax' : the values and indices

    Thus, to get maximum peak values at swing phase, use
    results['peaks']['swing']['max'].
    """

    if isinstance(curves, list):
        curves = np.array(curves)
    if isinstance(toeoffs, list):
        toeoffs = np.array(toeoffs)
    if curves.shape[0] != toeoffs.shape[0]:
        raise ValueError('invalid shape of arguments')
    # use defaultdict to reduct dict initialization boilerplate
    results = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
    results['contact'] = list()
    results['toeoff'] = list()

    for curve, toeoff in zip(curves, toeoffs):

        # extract stance and swing phases; swing phase begins at the toeoff frame
        curve_stance, curve_swing = np.split(curve, [toeoff])

        # get the simple extrema
        results['extrema']['overall']['min'].append(curve.min())
        results['extrema']['overall']['argmin'].append(curve.argmin())
        results['extrema']['overall']['max'].append(curve.max())
        results['extrema']['overall']['argmax'].append(curve.argmax())

        results['extrema']['stance']['min'].append(curve_stance.min())
        results['extrema']['stance']['argmin'].append(curve_stance.argmin())
        results['extrema']['stance']['max'].append(curve_stance.max())
        results['extrema']['stance']['argmax'].append(curve_stance.argmax())

        # swing phase indices need to be offset by the toeff frame
        results['extrema']['swing']['min'].append(curve_swing.min())
        results['extrema']['swing']['argmin'].append(curve_swing.argmin() + toeoff)
        results['extrema']['swing']['max'].append(curve_swing.max())
        results['extrema']['swing']['argmax'].append(curve_swing.argmax() + toeoff)

        # get the peaks (local extrema)
        ind, val = _get_local_min(curve)
        results['peaks']['overall']['argmin'].append(ind)
        results['peaks']['overall']['min'].append(val)
        ind, val = _get_local_max(curve)
        results['peaks']['overall']['argmax'].append(ind)
        results['peaks']['overall']['max'].append(val)

        ind, val = _get_local_min(curve_stance)
        results['peaks']['stance']['argmin'].append(ind)
        results['peaks']['stance']['min'].append(val)
        ind, val = _get_local_max(curve_stance)
        results['peaks']['stance']['argmax'].append(ind)
        results['peaks']['stance']['max'].append(val)

        ind, val = _get_local_min(curve_swing)
        results['peaks']['swing']['argmin'].append(ind + toeoff)
        results['peaks']['swing']['min'].append(val)
        ind, val = _get_local_max(curve_swing)
        results['peaks']['swing']['argmax'].append(ind + toeoff)
        results['peaks']['swing']['max'].append(val)

        # get some single values
        results['contact'].append(curve[0])
        results['toeoff'].append(curve[toeoff])

        # if needed, we can finally return a regular dict
        # results = dict(results)
        # for k in 'extrema', 'peaks':
        #     results[k] = dict(results[k])
        #     for x, v in results[k].items():
        #         results[k][x] = dict(v)

    return results


def _trials_extract_values(trials, from_models=None):
    """Extract curve values from given trials.

    Parameters
    ----------
    trials: list
        List of c3d files or Trial instances to extract data from.
    from_models : list
        List of GaitModel instances. These determine the variables to collect
        data for.

    Returns
    -------
    dict
        Dict keyed by variable. The values are dicts as returned by
        curve_extract_values().
    """
    if from_models is None:
        from_models = [
            models.pig_lowerbody,
        ]
    try:
        thevars = itertools.chain.from_iterable(mod.varnames for mod in from_models)
    except AttributeError:
        raise RuntimeError('')
    # collect all curves
    data, cycles = collect_trial_data(trials, collect_types=['model'])
    vals = dict()
    # extract values for each variable
    for var in thevars:
        data_var = data['model'][var]
        toeoffs_var = [cyc.toeoffn for cyc in cycles['model'][var]]
        if data_var is not None and toeoffs_var is not None:
            vals[var] = curve_extract_values(data_var, toeoffs_var)
        else:
            logger.info('no data for %s' % var)
    return vals



# -*- coding: utf-8 -*-
"""
Create gait reports in pdf format.

@author: Jussi (jnu@iki.fi)
"""

from __future__ import absolute_import
import itertools

import logging
import io
import os.path as op
from matplotlib.backends.backend_pdf import PdfPages
from matplotlib.figure import Figure
from collections import defaultdict, OrderedDict

from .. import sessionutils, normaldata, trial, models, stats
from ..envutils import GaitDataError
from ..config import cfg
from ..viz import timedist
from ulstools.num import age_from_hetu
from ..viz.plots import _plot_sessions, _plot_session_average, plot_trial_velocities
from ..viz.plot_matplotlib import _plot_extracted_table_plotly
from .text import _curve_extracted_text, _session_analysis_text
from .translations import translate


logger = logging.getLogger(__name__)

page_size = (11.69, 8.27)  # report page size = landscape A4
pdf_backend = 'matplotlib'


def _add_footer(fig, txt):
    """Add footer text to matplotlib Figure"""
    # XXX: currently puts text in right bottom corner
    fig.text(1, 0, txt, fontsize=8, color='black', ha='right', va='bottom')


def _add_header(fig, txt):
    """Add header text to matplotlib Figure"""
    # XXX: currently puts text in left bottom corner
    fig.text(0, 0, txt, fontsize=8, color='black', ha='left', va='bottom')


def _make_text_fig(txt, titlepage=True):
    """Make a Figure from given text.

    If titlepage is True, use larger font in bold.
    """
    fig = Figure()
    ax = fig.add_subplot(111)
    ax.set_axis_off()
    if titlepage:
        ax.text(0.5, 0.8, txt, ha='center', va='center', weight='bold', fontsize=14)
    else:
        ax.text(0.5, 0.8, txt, ha='center', va='center', fontsize=12)
    return fig


def _savefig(pdf, fig, header=None, footer=None):
    """Save figure fig into the given pdf object.

    Adds header/footer into page and saves as A4.
    """
    if fig is None:
        return
    elif not isinstance(fig, Figure):
        raise TypeError('fig must be matplotlib Figure, got %s' % fig)
    if header is not None:
        _add_header(fig, header)
    if footer is not None:
        _add_footer(fig, footer)
    fig.set_size_inches(page_size[0], page_size[1])
    pdf.savefig(fig)


def create_report(
    sessionpath,
    info=None,
    pages=None,
    destdir=None,
    write_timedist=False,
    write_extracted=False,
):
    """Create a single-session pdf report.

    Parameters
    ----------
    sessionpath : str
        Path to session.
    info : dict, optional
        Patient info dict.
    pages : dict, optional
        Which pages to include to include in report. Set value to True for
        desired pages. If None, do all pages. Currently supported keys:
        'TrialVelocity'
        'TimeDistAverage'
        'KinematicsCons'
        'KineticsCons'
        'MuscleLenCons'
        'KinAverage'
    destdir : str, optional
        Destination directory for the pdf report. If None, write into sessionpath.
    write_timedist : bool
        If True, also write a text report of time-distance parameters into the
        same directory as the pdf.
    write_extracted : bool
        If True, also write a text report of curve extracted values into the
        same directory as the pdf.

    Returns
    -------
    str
        A status message.
    """
    if info is None:
        info = defaultdict(lambda: '')
    fullname = info['fullname'] or ''
    hetu = info['hetu'] or ''
    session_description = info['session_description'] or ''
    if pages is None:
        pages = defaultdict(lambda: True)  # default: do all plots
    elif not any(pages.values()):
        pages = defaultdict(lambda: False)

    do_emg_consistency = True

    session_root, sessiondir = op.split(sessionpath)
    patient_code = op.split(session_root)[1]
    if destdir is None:
        destdir = sessionpath
    pdfname = sessiondir + '.pdf'
    pdfpath = op.join(destdir, pdfname)

    tagged_trials = sessionutils._get_tagged_dynamic_c3ds_from_sessions(
        [sessionpath], tags=cfg.eclipse.tags
    )
    # XXX: do we unnecessarily load all trials twice (since plotting calls below don't
    # use the trials loaded here), or is it a non issue due to caching of c3d files?
    trials = (trial.Trial(t) for t in tagged_trials)
    has_kinetics = any(c.on_forceplate for t in trials for c in t.cycles)

    session_t = sessionutils.get_session_date(sessionpath)
    logger.debug('session timestamp: %s', session_t)
    age = age_from_hetu(hetu, session_t) if hetu else None

    model_normaldata = normaldata._read_session_normaldata(sessionpath)

    # make header page
    # timestr = time.strftime('%d.%m.%Y')  # current time, not currently used
    title_txt = '%s\n' % cfg.report.laboratory_name
    title_txt += '%s\n' % translate('Results of gait analysis')
    title_txt += '\n'
    title_txt += u'%s: %s\n' % (translate('Name'), fullname)
    title_txt += u'%s: %s\n' % (
        translate('Social security number'),
        hetu if hetu else translate('unknown'),
    )
    age_str = '%d %s' % (age, translate('years')) if age else translate('unknown')
    title_txt += u'%s: %s\n' % (translate('Age at time of measurement'), age_str)
    title_txt += u'%s: %s\n' % (translate('Session'), sessiondir)
    if session_description:
        title_txt += u'%s: %s\n' % (translate('Description'), session_description)
    title_txt += u'%s: %s\n' % (
        translate('Session date'),
        session_t.strftime('%d.%m.%Y'),
    )
    title_txt += u'%s: %s\n' % (translate('Patient code'), patient_code)
    fig_title = _make_text_fig(title_txt)

    header = u'%s: %s %s: %s' % (
        translate('Name'),
        fullname,
        translate('Social security number'),
        hetu,
    )
    musclelen_ndata = normaldata._find_normaldata_for_age(age)
    footer_musclelen = (
        u' %s: %s' % (translate('Normal data'), musclelen_ndata)
        if musclelen_ndata
        else u''
    )

    # make the figures
    legend_type = cfg.report.legend_type  # not currently used (legends disabled)
    style_by = cfg.report.style_by
    color_by = cfg.report.color_by

    # trial velocity plot
    fig_vel = None
    if pages['TrialVelocity']:
        logger.debug('creating velocity plot')
        fig_vel = plot_trial_velocities(sessionpath, backend=pdf_backend)

    # time-distance average
    fig_timedist_avg = None
    if pages['TimeDistAverage']:
        logger.debug('creating time-distance plot')
        fig_timedist_avg = timedist.plot_session_average(
            sessionpath, backend=pdf_backend
        )

    # for next 2 plots, disable the legends (there are typically too many cycles)
    # kin consistency
    fig_kinematics_cons = None
    if pages['KinematicsCons']:
        logger.debug('creating kinematics consistency plot')
        fig_kinematics_cons = _plot_sessions(
            sessions=[sessionpath],
            layout_name='lb_kinematics',
            model_normaldata=model_normaldata,
            color_by=color_by,
            style_by=style_by,
            backend=pdf_backend,
            figtitle='Kinematics consistency for %s' % sessiondir,
            legend_type=legend_type,
            legend=False,
        )

    # kinetics consistency
    fig_kinetics_cons = None
    if pages['KineticsCons'] and has_kinetics:
        logger.debug('creating kinetics consistency plot')
        fig_kinetics_cons = _plot_sessions(
            sessions=[sessionpath],
            layout_name='lb_kinetics',
            model_normaldata=model_normaldata,
            color_by=color_by,
            style_by=style_by,
            backend=pdf_backend,
            figtitle='Kinetics consistency for %s' % sessiondir,
            legend_type=legend_type,
            legend=False,
        )

    # musclelen consistency
    fig_musclelen_cons = None
    if pages['MuscleLenCons']:
        logger.debug('creating muscle length consistency plot')
        fig_musclelen_cons = _plot_sessions(
            sessions=[sessionpath],
            layout_name='musclelen',
            color_by=color_by,
            style_by=style_by,
            model_normaldata=model_normaldata,
            backend=pdf_backend,
            figtitle='Muscle length consistency for %s' % sessiondir,
            legend_type=legend_type,
            legend=False,
        )
    # EMG consistency
    fig_emg_cons = None
    if do_emg_consistency:
        logger.debug('creating EMG consistency plot')
        fig_emg_cons = _plot_sessions(
            sessions=[sessionpath],
            layout_name='std_emg',
            color_by=color_by,
            style_by=style_by,
            figtitle='EMG consistency for %s' % sessiondir,
            legend_type=legend_type,
            legend=False,
            backend=pdf_backend,
        )
    # average plots, R/L
    fig_kin_avg = None
    if pages['KinAverage']:
        fig_kin_avg = _plot_session_average(
            sessionpath,
            model_normaldata=model_normaldata,
            backend=pdf_backend,
        )

    # prep for extracted values if needed
    if pages['Extracted'] or write_extracted:
        vardefs_dict = OrderedDict(cfg.report.vardefs)
        allvars = [vardef[0] for vardefs in vardefs_dict.values() for vardef in vardefs]
        from_models = set(models.model_from_var(var) for var in allvars)
        curve_vals = {
            op.split(sessionpath)[-1]: stats._trials_extract_values(
                tagged_trials, from_models=from_models
            )
        }

    # tables of curve extracted values
    figs_extracted = list()
    if pages['Extracted']:
        logger.debug('plotting curve extracted values')
        for title, vardefs in vardefs_dict.items():
            fig = _plot_extracted_table_plotly(curve_vals, vardefs)
            fig.tight_layout()
            fig.set_dpi(300)
            fig.suptitle('Curve extracted values: %s' % title)
            figs_extracted.append(fig)

    # save the pdf file
    logger.debug('creating multipage pdf %s' % pdfpath)
    with PdfPages(pdfpath) as pdf:
        _savefig(pdf, fig_title)
        _savefig(pdf, fig_vel, header)
        _savefig(pdf, fig_timedist_avg, header)
        _savefig(pdf, fig_kinematics_cons, header)
        _savefig(pdf, fig_kinetics_cons, header)
        _savefig(pdf, fig_musclelen_cons, header, footer_musclelen)
        _savefig(pdf, fig_emg_cons, header)
        _savefig(pdf, fig_kin_avg, header)
        for fig in figs_extracted:
            _savefig(pdf, fig, header)

    # save the time-distance parameters into a text file
    if write_timedist:
        _timedist_txt = _session_analysis_text(sessionpath)
        timedist_txt_file = sessiondir + '_time_distance.txt'
        timedist_txt_path = op.join(destdir, timedist_txt_file)
        with io.open(timedist_txt_path, 'w', encoding='utf8') as f:
            logger.debug('writing timedist text data into %s' % timedist_txt_path)
            f.write(_timedist_txt)

    # save the curve extraced values into a text file
    if write_extracted:
        extracted_txt = '\n'.join(_curve_extracted_text(curve_vals, vardefs_dict))
        extracted_txt_file = sessiondir + '_curve_values.txt'
        extracted_txt_path = op.join(destdir, extracted_txt_file)
        with io.open(extracted_txt_path, 'w', encoding='utf8') as f:
            logger.debug('writing extracted text data into %s' % extracted_txt_path)
            f.write(extracted_txt)

    return 'Created %s' % pdfpath


def create_comparison_report(
    sessionpaths, info=None, pages=None, destdir=None, write_extracted=False
):
    """Create a comparison pdf report.

    Parameters
    ----------
    sessionpaths : list
        List of paths to sessions.
    info : dict, optional
        Patient info dict.
    pages : dict, optional
        Which pages to include to include in report. Set value to True for
        desired pages. Currently supported keys:
        'TimeDist'
        'Kinematics'
        'Kinetics'
        'MuscleLen'
        'Extracted'
    destdir : str, optional
        Destination directory for the pdf report. If None, write into first
        session path.

    Returns
    -------
    str
        A status message.
    """

    if len(sessionpaths) <= 1:
        raise GaitDataError('need at least two sessions')

    if info is None:
        info = defaultdict(lambda: '')

    if pages is None:
        # if no pages specified, do them all
        pages = defaultdict(lambda: True)
    elif not any(pages.values()):
        pages = defaultdict(lambda: False)

    # gather trials and check for kinetics
    trials_dict = OrderedDict()  # Py2: preserve session ordering
    for session in sessionpaths:
        trials_dict[session] = sessionutils._get_tagged_dynamic_c3ds_from_sessions(
            session, tags=cfg.eclipse.tags
        )
    alltrials = (
        trial.Trial(t) for t in itertools.chain.from_iterable(trials_dict.values())
    )
    any_kinetics = any(c.on_forceplate for t in alltrials for c in t.cycles)

    # compose a name for the resulting pdf; it will be saved in the first session dir
    sessionpath = sessionpaths[0]
    if destdir is None:
        destdir = sessionpath
    # XXX: filenames can become very long here
    pdfname = ' VS '.join(op.split(sp)[1] for sp in sessionpaths) + '.pdf'
    pdfpath = op.join(destdir, pdfname)

    # read model normaldata according to first session
    # age may be different for different sessions but this is probably good enough
    model_normaldata = normaldata._read_session_normaldata(sessionpath)

    # make header page
    fullname = info['fullname'] or ''
    hetu = info['hetu'] or ''
    title_txt = '%s\n' % cfg.report.laboratory_name
    title_txt += u'%s\n' % translate('Comparison report')
    title_txt += '\n'
    title_txt += info['session_description']
    title_txt += '\n'
    title_txt += u'%s: %s\n' % (translate('Name'), fullname)
    title_txt += u'%s: %s\n' % (translate('Social security number'), hetu)
    fig_title = _make_text_fig(title_txt)

    # make the figures
    legend_type = cfg.report.comparison_legend_type
    style_by = cfg.report.comparison_style_by
    color_by = cfg.report.comparison_color_by
    emg_mode = 'rms' if cfg.report.comparison_emg_rms else None

    fig_timedist = (
        timedist.plot_comparison(sessionpaths, backend=pdf_backend)
        if pages['TimeDist']
        else None
    )

    fig_kinematics = (
        _plot_sessions(
            sessionpaths,
            tags=cfg.eclipse.repr_tags,
            layout_name='lb_kinematics',
            model_normaldata=model_normaldata,
            figtitle='Kinematics comparison',
            style_by=style_by,
            color_by=color_by,
            backend=pdf_backend,
            legend_type=legend_type,
        )
        if pages['Kinematics']
        else None
    )

    fig_kinetics = (
        _plot_sessions(
            sessionpaths,
            tags=cfg.eclipse.repr_tags,
            layout_name='lb_kinetics',
            model_normaldata=model_normaldata,
            figtitle='Kinetics comparison',
            style_by=style_by,
            color_by=color_by,
            legend_type=legend_type,
            backend=pdf_backend,
        )
        if pages['Kinetics'] and any_kinetics
        else None
    )

    fig_musclelen = (
        _plot_sessions(
            sessionpaths,
            tags=cfg.eclipse.repr_tags,
            layout_name='musclelen',
            model_normaldata=model_normaldata,
            figtitle='Muscle length comparison',
            style_by=style_by,
            color_by=color_by,
            legend_type=legend_type,
            backend=pdf_backend,
        )
        if pages['MuscleLen']
        else None
    )

    fig_emg = (
        _plot_sessions(
            sessionpaths,
            tags=cfg.eclipse.repr_tags,
            layout_name='std_emg',
            emg_mode=emg_mode,
            figtitle='EMG comparison',
            style_by=style_by,
            color_by=color_by,
            legend_type=legend_type,
            backend=pdf_backend,
        )
        if pages['MuscleLen']
        else None
    )

    # prep for extracted values if needed
    if pages['Extracted'] or write_extracted:
        vardefs_dict = OrderedDict(cfg.report.vardefs)
        allvars = [vardef[0] for vardefs in vardefs_dict.values() for vardef in vardefs]
        from_models = set(models.model_from_var(var) for var in allvars)
        curve_vals = OrderedDict(
            [
                (
                    op.split(session)[-1],
                    stats._trials_extract_values(trials, from_models=from_models),
                )
                for session, trials in trials_dict.items()
            ]
        )

    # tables of curve extracted values
    figs_extracted = list()
    if pages['Extracted']:
        logger.debug('plotting curve extracted values')
        for title, vardefs in vardefs_dict.items():
            fig = _plot_extracted_table_plotly(curve_vals, vardefs)
            fig.tight_layout()
            fig.set_dpi(300)
            fig.suptitle('Curve extracted values: %s' % title)
            figs_extracted.append(fig)

    header = u'%s: %s %s: %s' % (
        translate('Name'),
        fullname,
        translate('Social security number'),
        hetu,
    )
    logger.debug('creating multipage comparison pdf %s' % pdfpath)
    with PdfPages(pdfpath) as pdf:
        _savefig(pdf, fig_title)
        _savefig(pdf, fig_timedist, header)
        _savefig(pdf, fig_kinematics, header)
        _savefig(pdf, fig_kinetics, header)
        _savefig(pdf, fig_musclelen, header)
        _savefig(pdf, fig_emg, header)
        for fig in figs_extracted:
            _savefig(pdf, fig, header)

    # save the curve extraced values into a text file
    if write_extracted:
        extracted_txt = '\n'.join(_curve_extracted_text(curve_vals, vardefs_dict))
        extracted_txt_file = ' VS '.join(op.split(sp)[1] for sp in sessionpaths)
        extracted_txt_file += '_curve_values.txt'
        extracted_txt_path = op.join(destdir, extracted_txt_file)
        with io.open(extracted_txt_path, 'w', encoding='utf8') as f:
            logger.debug('writing extracted text data into %s' % extracted_txt_path)
            f.write(extracted_txt)

    return 'Created %s\nin %s' % (pdfname, sessionpath)



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 23 11:17:10 2016

Misc numerical utils

@author: Jussi (jnu@iki.fi)
"""

from __future__ import division

import logging
import numpy as np
import hashlib
from scipy import signal
from scipy.signal import medfilt
from scipy.special import erfcinv

logger = logging.getLogger(__name__)


def kabsch_rotation(P, Q):
    """Calculate a rotation matrix P->Q using the Kabsch algorithm."""
    H = np.dot(P.T, Q)
    U, S, V = np.linalg.svd(H)
    E = np.eye(3)
    E[2, 2] = np.sign(np.linalg.det(np.dot(V.T, U.T)))
    return np.dot(np.dot(V.T, E), U.T)


def _rotation_matrix(yaw, pitch, roll):
    """Rotation matrix from yaw, pitch, roll in degrees"""
    yaw, pitch, roll = np.array([yaw, pitch, roll]) / 180 * np.pi
    cos, sin = np.cos, np.sin
    r1 = [
        cos(yaw) * cos(pitch),
        cos(yaw) * sin(pitch) * sin(roll) - sin(yaw) * cos(roll),
        cos(yaw) * sin(pitch) * cos(roll) + sin(yaw) * sin(roll),
    ]
    r2 = [
        sin(yaw) * cos(pitch),
        sin(yaw) * sin(pitch) * sin(roll) + cos(yaw) * cos(roll),
        sin(yaw) * sin(pitch) * cos(roll) - cos(yaw) * sin(roll),
    ]
    r3 = [-sin(pitch), cos(pitch) * sin(roll), cos(pitch) * cos(roll)]
    return np.array([r1, r2, r3])


def _rigid_body_extrapolate_markers(P0, Pr):
    """Extrapolate some markers in a rigid set.
    P0 (N x 3), the marker positions in the static frame.
    Pr (M x 3), the marker positions in where extrapolation is needed.
    N-M markers will be extrapolated.

    Algorithm:
    -find R and t that take us from Pr0 to Pr, by Kabsch algorithm (Pr0 is the
    reference markers in P0)
    -apply R and t to P0 to get the extrapolated positions
    """
    nref = Pr.shape[0]
    if P0.shape <= nref:
        raise ValueError('1st dim of P0 needs to be larger than 1st dim of Pr')
    Pr0 = P0[:nref, :]  # reference markers
    # find rotation and translation that take the static reference position to the
    # position where extrapolation is needed
    trans0 = Pr0.mean(axis=0)
    Pr0_ = Pr0 - trans0
    P0_ = P0 - trans0
    trans1 = Pr.mean(axis=0)
    Pr_ = Pr - trans1
    R = kabsch_rotation(Pr0_, Pr_)
    # apply the transformation to the markers to be extrapolated
    return np.dot(R, P0_[nref:, :].T).T + trans1


def mad(data, axis=None, scale=1.4826, keepdims=False):
    """Median absolute deviation (MAD).

    Defined as the median absolute deviation from the median of the data. A
    robust alternative to stddev. Results should be identical to
    scipy.stats.median_absolute_deviation(), which does not take a keepdims
    argument.

    Parameters
    ----------
    data : array_like
        The data.
    scale : float, optional
        Scaling of the result. By default, it is scaled to give a consistent
        estimate of the standard deviation of values from a normal
        distribution.
    axis : numpy axis spec, optional
        Axis or axes along which to compute MAD.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left in the
        result as dimensions with size one.

    Returns
    -------
    ndarray
        The MAD.
    """
    # keep dims here so that broadcasting works
    med = np.median(data, axis=axis, keepdims=True)
    abs_devs = np.abs(data - med)
    return scale * np.median(abs_devs, axis=axis, keepdims=keepdims)


def modified_zscore(data, axis=None, single_mad=None):
    """Modified Z-score.

    Z-score analogue computed using robust (median-based) statistics.

    Parameters
    ----------
    data : array_like
        The data
    axis : numpy axis spec, optional
        Axis or axes along which to compute the statistic.
    single_mad : bool, optional
        Use a single MAD estimate computed all over the data. If False, MAD
        will be computed along given axis (e.g. separately for each variable).

    Returns
    -------
    ndarray
        The modified Z-score.
    """
    med_ = np.median(data, axis=axis, keepdims=True)
    mad_ = mad(data, axis=axis, keepdims=True)
    if single_mad:
        mad_ = np.median(mad_)
    return (data - med_) / mad_


def outliers(x, axis=0, single_mad=None, p_threshold=1e-3):
    """Robustly detect outliers assuming a normal distribution.

    A modified Z-score is first computed based on the data. Then a threshold
    value Zlim is computed from p_threshold, and values that exceed Zlim are
    rejected. p_threshold is the probability of rejection assuming strictly
    normally distributed data, i.e. probability for "false outlier"

    Parameters
    ----------
    data : array_like
        The data.
    axis : numpy axis spec, optional
        Axis or axes along which to compute the Z scores. E.g. axis=0
        computes row-wise Z scores and rejects based on those.
    single_mad : bool
        Use a single MAD estimate computed all over the data. If False, the MAD
        will be computed along given axis (e.g. separately for each variable).
    p_threshold : float
        The P threshold for rejection.

    Returns
    -------
    tuple
        Indexes of rejected values (np.where output)
    """
    zs = modified_zscore(x, axis=axis, single_mad=single_mad)
    z_threshold = np.sqrt(2) * erfcinv(p_threshold)
    logger.debug('Z threshold: %.2f' % z_threshold)
    return np.where(abs(zs) > z_threshold)


def _files_digest(files):
    """Create overall md5 digest for a sequence of files (filenames)"""
    hashes = sorted(_file_digest(fn) for fn in files)
    # concat as unicode and encode to get a well defined byte representation in
    # both py2 and py3
    hash_str = u''.join(hashes).encode('utf-8')
    return hashlib.md5(hash_str).hexdigest()


def _file_digest(fn):
    """Return md5 digest for file"""
    with open(fn, 'rb') as f:
        data = f.read()
    return hashlib.md5(data).hexdigest()


def rising_zerocross(x):
    """Find indices of rising zero crossings in array.

    The indices are defined as n for which x[n] >= 0 and x[n-1] < 0.

    Parameters
    ----------
    x : array_like
        The data.

    Returns
    -------
    tuple
        The indices (np.where output)
    """
    x = np.array(x)  # we can also handle lists etc.
    return np.where(np.logical_and(x[1:] >= 0, x[:-1] < 0))[0] + 1


def falling_zerocross(x):
    """Find indices of falling zero crossings in array.

    The indices are defined as n for which x[n] <= 0 and x[n-1] > 0.

    Parameters
    ----------
    x : array_like
        The data.

    Returns
    -------
    tuple
        The indices (np.where output)
    """
    return rising_zerocross(-x)


def digitize_array(v, b):
    """Replace all elements of v with their closest matches in b.

    Uses abs(x-y) as the distance metric. This function is similar to
    np.digitize(), but simpler in usage and implementation.

    Parameters
    ----------
    v : array_like
        Array to make replacements in.
    b : array_like
        Array to pick replacements from.

    Returns
    -------
    ndarray
        The result.
    """
    v = np.array(v)
    b = np.array(b)
    if b.size == 0:
        return v
    inds = np.abs(v[np.newaxis, :] - b[:, np.newaxis]).argmin(axis=0)
    return b[inds]


def _isfloat(x):
    """Test for float-conversible value"""
    try:
        float(x)
        return True
    except ValueError:
        return False


def _is_ascii(s):
    """Check for ASCII string"""
    return all(ord(c) < 128 for c in s)


def _isint(x):
    """Test for int-conversible value"""
    try:
        int(x)
        return True
    except ValueError:
        return False


def _baseline(v):
    """Baseline v using histogram.

    Subtracts the most prominent signal level.
    """
    v = np.array(v)
    v = v.squeeze()
    if len(v.shape) != 1:
        raise ValueError('Need 1-dim input')
    nbins = max(int(len(v) / 10), 1000)  # exact n of bins should not matter
    ns, edges = np.histogram(v, bins=nbins)
    peak_ind = np.where(ns == np.max(ns))[0][0]
    return v - np.mean(edges[peak_ind : peak_ind + 2])


def center_of_pressure(F, M, dz):
    """Compute center of pressure from forceplate data.

    Computes CoP according to AMTI formula. The results differ slightly
    (typically few mm) from those given by Nexus, for unknown reasons (different
    filter?) See http://health.uottawa.ca/biomech/courses/apa6903/amticalc.pdf

    Parameters
    ----------
    F : ndarray
        The force vector (Nx3). N is the number of observations.
    M : ndarray
        The moment vector (Nx3)
    dz : float
        The thickness of the plate, or the distance from moment origin to physical origin.

    Returns
    -------
    ndarray
        The center of pressure (Nx3)
    """
    FP_FILTFUN = medfilt  # the filter function
    FP_FILTW = 5  # median filter width
    fx, fy, fz = tuple(F.T)  # split columns into separate vars
    mx, my, _ = tuple(M.T)
    fz = FP_FILTFUN(fz, FP_FILTW)
    nz_inds = np.where(np.abs(fz) > 0)[0]  # only divide on nonzero inds
    cop = np.zeros((fx.shape[0], 3))
    cop[nz_inds, 0] = -(my[nz_inds] + fx[nz_inds] * dz) / fz[nz_inds]
    cop[nz_inds, 1] = (mx[nz_inds] - fy[nz_inds] * dz) / fz[nz_inds]
    return cop


def _change_coords(pts, wR, wT):
    """Translate pts (Nx3) into a new coordinate system described by
    rotation matrix wR and translation vector wT"""
    pts = np.array(pts)
    return np.dot(wR, pts.T).T + wT


def _segment_angles(P):
    """Compute angles between line segments.

    The segments are defined by ordered points in P (Nx3 array). For N points,
    there will be N-1 segments and N-2 angles between those. It can also be 3-d
    matrix of (TxNx3) to get time-dependent data. Output will be (N-2) vector or
    Tx(N-2) matrix of angles in radians. If successive points are identical,
    nan:s will be output for the corresponding angles.
    """
    if P.shape[-1] != 3 or len(P.shape) not in [2, 3]:
        raise ValueError('Invalid shape of input matrix')
    if len(P.shape) == 2:
        P = P[np.newaxis, ...]  # insert singleton time axis
    Pd = np.diff(P, axis=1)  # point-to-point vectors
    vnorms = np.linalg.norm(Pd, axis=2)[..., np.newaxis]
    # ignore 0/0 and x/0 errors -> nan
    with np.errstate(divide='ignore', invalid='ignore'):
        Pdn = Pd / vnorms
    # take dot products between successive vectors and angles by arccos
    dots = np.sum(Pdn[:, 0:-1, :] * Pdn[:, 1:, :], axis=2)
    dots = dots[0, :] if dots.shape[0] == 1 else dots  # rm singleton dim
    return np.pi - np.arccos(dots)


def _running_sum(M, win, axis=None):
    """Calculate running sum for given window length.

    M is the data (ndarray) and sum will be along given axis. win is the window
    length. Uses cumulative sum, inspired by:
    http://arogozhnikov.github.io/2015/09/30/NumpyTipsAndTricks2.html"""
    if axis is None:
        M = M.flatten()
    s = np.cumsum(M, axis=axis)
    s = np.insert(s, 0, [0], axis=axis)
    len_ = s.shape[0] if axis is None else s.shape[axis]
    return s.take(np.arange(win, len_), axis=axis) - s.take(
        np.arange(0, len_ - win), axis=axis
    )


def rms(data, win, axis=None, pad_mode=None):
    """Calculate rolling window RMS.

    Parameters
    ----------
    data : ndarray
        The data.
    win : int
        Window length.
    axis : numpy axis spec, optional
        Axis along which to compute rms.
    pad_mode : str or function, optional
        Padding mode. See np.pad for details.

    Returns
    -------
    ndarray
        The rms data.
    """
    if pad_mode is None:
        pad_mode = 'edge'
    if win % 2 != 1:
        raise ValueError('Need RMS window of odd length')
    datalen = len(data) if axis is None else data.shape[axis]
    if win > datalen:
        raise ValueError('Need win length < data length')
    rms_ = np.sqrt(_running_sum(data ** 2, win, axis=axis) / win)
    # pad RMS data so that lengths are matched
    padw = int((win - 1) / 2)
    padarg_axis = (padw, padw)
    if axis == None:
        eff_dim = 1
        axis = 0
    else:
        eff_dim = data.ndim
    padarg = eff_dim * [(0, 0)]
    padarg[axis] = padarg_axis
    return np.pad(rms_, padarg, mode=pad_mode)


def _filtfilt(data, passband, sfrate, buttord=5):
    """Forward-backward filter.
    Filter data into given passband, e.g. [1, 40].
    Frequencies are given in Hz along with sfrate (sampling rate).
    Implemented as pure lowpass, if highpass freq = 0.
    """
    if passband is None:
        return data
    passbandn = 2 * np.array(passband) / sfrate
    if passbandn[0] > 0:  # bandpass
        b, a = signal.butter(buttord, passbandn, 'bandpass')
    else:  # lowpass
        b, a = signal.butter(buttord, passbandn[1])
    return signal.filtfilt(b, a, data)


def _get_local_max(data):
    """Get local maximum (peak) of 1-D data"""
    # the simpler argrelextrema() would not return flat peaks, which might occur
    # at least in theory
    peak_inds = signal.find_peaks(data)[0]
    if not any(peak_inds):
        return (np.nan, np.nan)
    peak_ind = peak_inds[data[peak_inds].argmax()]
    peak_val = data[peak_ind]
    return peak_ind, peak_val


def _get_local_min(data):
    """Get local minimum  (peak) of 1-D data"""
    ind, val = _get_local_max(-data)
    if ind is np.nan:
        return np.nan, np.nan
    else:
        return ind, -val



# -*- coding: utf-8 -*-
"""

Console script wrappers for the plotters. These are used as setuptools
entry points.

@author: Jussi (jnu@iki.fi)
"""


import logging
import argparse

from . import plots
from .plot_misc import show_fig
from ..envutils import _register_gui_exception_handler
from .. import nexus


def _console_init():
    """Set up some things for console scripts"""
    logging.basicConfig(level=logging.DEBUG)
    _register_gui_exception_handler()


def plot_nexus_trial():
    """Plot currently loaded Nexus trial"""
    _console_init()
    parser = argparse.ArgumentParser()
    parser.add_argument('--layout', type=str)
    parser.add_argument('--backend', type=str)
    args = parser.parse_args()
    fig = plots.plot_nexus_trial(layout_name=args.layout, backend=args.backend)
    show_fig(fig)


def plot_nexus_session():
    """Plot tagged dynamic trials from current Nexus session"""
    _console_init()
    parser = argparse.ArgumentParser()
    parser.add_argument('--layout', type=str)
    parser.add_argument('--backend', type=str)
    parser.add_argument('--tags', nargs='+')
    args = parser.parse_args()
    sessions = [nexus.get_sessionpath()]
    fig = plots._plot_sessions(
        sessions, tags=args.tags, backend=args.backend, layout_name=args.layout
    )
    show_fig(fig)


def plot_nexus_session_average():
    """Kin average plot for session"""
    _console_init()
    parser = argparse.ArgumentParser()
    parser.add_argument('--layout', type=str)
    parser.add_argument('--backend', type=str)
    args = parser.parse_args()
    session = nexus.get_sessionpath()
    fig = plots._plot_session_average(
        session, layout_name=args.layout, backend=args.backend
    )
    show_fig(fig)



# -*- coding: utf-8 -*-
"""
Wrapper methods that read from Vicon Nexus or c3d files.
The 'source' argument can be either a ViconNexus.ViconNexus instance or a c3d
filename. Returned values should be independent of source.

@author: Jussi (jnu@iki.fi)
"""
from __future__ import division

import numpy as np
import logging

from . import nexus, c3d
from .config import cfg


logger = logging.getLogger(__name__)


def _reader_module(source):
    """Determine the appropriate data reader module to use"""
    if nexus._is_vicon_instance(source):
        return nexus
    elif c3d._is_c3d_file(source):
        return c3d
    else:
        raise RuntimeError('Unknown type for data source %s' % source)


def get_metadata(source):
    """Get trial metadata from a source.

    Parameters
    ----------
    source : ViconNexus | str
        The data source. Can be a c3d filename or a ViconNexus instance.

    Returns
    -------
    dict
        Metadata dict with the following keys and values:

        trialname : str
            Name of trial.
        eclipse_data : dict
            The Eclipse data for the trial. Keys are Eclipse fields and values are
            the corresponding data.
        sessionpath : str
            Full path to session directory.
        length : int
            Trial length in frames.
        offset : int
            Frame offset of data from beginning of trial. The event numbers will be
            interpreted relative to the offset, e.g. if the trial offset is 100 and
            an event occurs at frame 150, its index in the frame data array is 50.
        framerate : float
            Frame rate for capture (frames / sec).
        analograte : float
            Sampling rate for analog devices (samples / sec)-
        name : str
            Subject name.
        subj_params : dict
            Other subject parameters (bodymass etc.)
        events : TrialEvents
            Trial events (foot strikes, toeoffs etc.)
    """
    meta = _reader_module(source)._get_metadata(source)
    # Nexus uses slightly different metadata field names as c3d,
    # so translate here into c3d naming convention
    # XXX: c3d angle params are apparently in radians while Nexus uses degrees
    # - NOT translated here!
    if nexus._is_vicon_instance(source):
        pars = meta['subj_params']
        for par in pars.keys():
            if 'Right' in par:
                new_par = par.replace('Right', 'R')
                pars[new_par] = pars.pop(par)
            if 'Left' in par:
                new_par = par.replace('Left', 'L')
                pars[new_par] = pars.pop(par)
    return meta


def get_forceplate_data(source):
    """Get forceplate data.

    Parameters
    ----------
    source : ViconNexus | str
        The data source. Can be a c3d filename or a ViconNexus instance.

    Returns
    -------
    dict
        The forceplate data dict with the following keys and values:
        F : ndarray
            Nx3 array with x,y,z components of the force.
        M : ndarray
            Nx3 array with x,y,z components of the moment.
        Ftot : ndarray
            Nx1 array of total force.
        CoP : ndarray
            Nx3 array of center of pressure data.
        analograte : float
            Sampling rate.
        samplesperframe : int
            Analog samples per capture frame.
    """
    return _reader_module(source)._get_forceplate_data(source)


def get_marker_data(source, markers, ignore_missing=False):
    """Get position, velocity and acceleration for given markers.

    Parameters
    ----------
    source : ViconNexus | str
        The data source. Can be a c3d filename or a ViconNexus instance.
    markers : list | str
        Marker name, or list of marker names.
    ignore_missing : bool, optional
        If True, ignore missing markers on read. Otherwise raise an exception.

    Returns
    -------
    dict
        Marker data dict. Keys are marker names and values are Nx3 ndarrays of
        x,y,z data.
    """
    return _reader_module(source)._get_marker_data(
        source,
        markers,
        ignore_missing=ignore_missing,
    )


def get_emg_data(source):
    """Read EMG data.

    Parameters
    ----------
    source : ViconNexus | str
        The data source. Can be a c3d filename or a ViconNexus instance.

    Returns
    -------
    dict
        Dict with following keys:
        t : ndarray
            Time axis in seconds.
        data : dict
            The data. Keys are channel names and values are ndarrays.
    """
    return _reader_module(source)._get_emg_data(source)


def get_analysis(source, condition='unknown'):
    """Read analysis data (e.g. time-distance vars).

    Parameters
    ----------
    source : str
        Name of a c3d file. Reads from Nexus are not supported yet.
    condition : str, optional
        The condition name for the analysis dict, by default 'unknown'.

    Returns
    -------
    dict
        A nested dict of the analysis values, keyed by variable name and
        context. The first key is the condition name.
    """
    if nexus._is_vicon_instance(source):
        raise Exception('Analysis var reads from Nexus not supported yet')
    return _reader_module(source).get_analysis(source, condition)


def get_accelerometer_data(source):
    """Read accelerometer data.

    Parameters
    ----------
    source : ViconNexus | str
        The data source. Can be a c3d filename or a ViconNexus instance.

    Returns
    -------
    dict
        Dict with following keys:
        t : ndarray
            Time axis in seconds.
        data : dict
            The data. Keys are channel names and values are ndarrays.
    """
    return _reader_module(source)._get_accelerometer_data(source)


def get_model_data(source, model):
    """Read model data (e.g. Plug-in Gait).

    Parameters
    ----------
    source : ViconNexus | str
        The data source. Can be a c3d filename or a ViconNexus instance.
    model : GaitModel
        The model to read. For available models, see models.py. For a known
        variable name, the corresponding model can be obtained by calling
        models.model_from_var(varname).

    Returns
    -------
    dict
        The model data. Keys are model variable names and values are ndarrays of
        data.
    """
    modeldata = _reader_module(source)._get_model_data(source, model)
    for var in model.read_vars:
        # convert Moment variables into SI units
        if var.find('Moment') > 0:
            modeldata[var] /= 1.0e3  # Nmm -> Nm
        # split 3D arrays into x,y,z variables
        if model.read_strategy == 'split_xyz':
            if modeldata[var].shape[0] == 3:
                modeldata[var + 'X'] = modeldata[var][0, :]
                modeldata[var + 'Y'] = modeldata[var][1, :]
                modeldata[var + 'Z'] = modeldata[var][2, :]
            else:
                raise RuntimeError('Expected a 3D array')
    # For basic Plug-in Gait, add the tibial torsion value to knee rotation.
    # This is to compensate for a weird PiG feature that offsets knee rotation
    # by tibial torsion, so that it always rotates around zero. With this fix,
    # we get the actual (anatomical) knee rotation.
    if (
        model.desc == 'Plug-in Gait lower body kinematics'
        and cfg.models.add_tibial_torsion
    ):
        params = get_metadata(source)['subj_params']
        for ctxt in 'RL':
            var_knee = ctxt + 'KneeAnglesZ'
            var_torsion = ctxt + 'TibialTorsion'
            if var_torsion in params and var_knee in modeldata:
                tibt = params[var_torsion]
                # for c3d data, need to convert radians -> degrees
                if c3d._is_c3d_file(source):
                    tibt /= np.pi / 180
                if np.abs(tibt) > 1e-2:  # do not add insignificant values
                    logger.info('adding %s tibial torsion: %g deg' % (ctxt, tibt))
                    modeldata[var_knee] += tibt
    return modeldata



# -*- coding: utf-8 -*-
"""

Definitions for various gait related models (Plug-in Gait etc.)

To create a new model, create a GaitModel() instance, fill in the data and
append to models_all.

@author: Jussi (jnu@iki.fi)
"""

from past.builtins import basestring
from builtins import object
from collections import defaultdict

# list of all our models
models_all = list()


def model_from_var(var_):
    """Return model corresponding to a variable.

    Parameters
    ----------
    var_ : str
        The variable name.

    Returns
    -------
    GaitModel | None
        The model. None if no models contain the given variable.
    """
    if var_ is None:
        return None
    elif not isinstance(var_, basestring):
        raise TypeError('Variable name must be a string or None')
    for model in models_all:
        if var_ in model.varnames or var_ in model.varnames_noside:
            return model
    return None


# convenience methods for model creation


def _list_with_side(vars_):
    """Prepend variable names in vars_ with 'L' and 'R'."""
    return [side + var for var in vars_ for side in 'LR']


def _dict_with_side_gen(di, append_side=False):
    """Prepend all dict keys in di with 'R' and 'L' and return new dict.
    If append_side, also append corresponding ' (R)' or ' (L)' to every dict
    value."""
    for key, val in di.items():
        for ctxt in 'RL':
            if append_side:
                yield ctxt + key, '%s (%s)' % (val, ctxt)
            else:
                yield ctxt + key, val


def _dict_with_side(di, append_side=False):
    """Prepend all dict keys in di with 'R' and 'L' and return new dict.
    If append_side, also append corresponding ' (R)' or ' (L)' to every dict
    value."""
    return dict(_dict_with_side_gen(di, append_side=append_side))


class GaitModel(object):
    """A class for storing information about a gait model.

    The data describes model-specific variable names etc. and is intended
    (currently not forced) to be non-mutable. The actual data is stored
    elsewhere.
    """

    def __init__(self):
        self.read_vars = list()  # vars to be read from data
        # How to read multidimensional variables: 'split_xyz' splits each
        # variable into x,y,z components and names them appropriately.
        # 'last' reads the last component only (c3d scalars are saved
        # as last component of 3-d array (??))
        self.read_strategy = None
        self.desc = ''  # description of model
        self.varnames = list()  # resulting variable names
        self.varnames_noside = list()  # variables without side
        self.varlabels_noside = dict()  # variables without side
        self.varlabels = dict()  # descriptive label for each variable
        # mapping from variable names to gcd normal data variables
        self.gcd_normaldata_map = dict()
        # units for the variables
        self.units = dict()
        # description of neg/pos dirs for each variable in
        # tuples, e.g. ('Plantarflexion', 'Dorsiflexion')
        self.ydesc = dict()
        # callable to return whether var requires valid forceplate contact
        self.is_kinetic_var = lambda var: False


"""Create the models"""

#
# Oxford foot model
#
ofm = GaitModel()
ofm.desc = 'Oxford foot model kinematics'
ofm.read_strategy = 'split_xyz'

ofm.read_vars = _list_with_side(['FFHFA', 'FFTBA', 'HFTBA', 'HFTFL', 'HXFFA', 'TIBA'])

ofm.varlabels_noside = {
    'FFHFAX': 'Forefoot-hindfoot dorsiflexion',
    'FFHFAY': 'Forefoot-hindfoot adduction',
    'FFHFAZ': 'Forefoot-hindfoot supination',
    'FFTBAX': 'Forefoot-tibia dorsiflexion',
    'FFTBAY': 'Forefoot-tibia adduction',
    'FFTBAZ': 'Forefoot-tibia supination',
    'HFTBAX': 'Hindfoot-tibia dorsiflexion',
    'HFTBAY': 'Hindfoot-tibia rotation',
    'HFTBAZ': 'Hindfoot-tibia inversion',
    'HFTFLX': 'Hindfoot-lab dorsiflexion',
    'HFTFLY': 'Hindfoot-lab rotation',
    'HFTFLZ': 'Hindfoot-lab inversion',
    'HXFFAX': 'Hallux-forefoot dorsiflexion',
    'HXFFAY': 'Hallux-forefoot varus',
    'HXFFAZ': 'NA',
    'TIBAX': 'Tibia-lab forward tilt',
    'TIBAY': 'Tibia-lab lateral tilt',
    'TIBAZ': 'Tibia internal rotation',
}

ofm.units = defaultdict(lambda: 'deg')
ofm.ydesc = _dict_with_side(
    {
        'FFHFAX': ('Plantarflexion', 'Dorsiflexion'),
        'FFHFAY': ('Abduction', 'Adduction'),
        'FFHFAZ': ('Pronation', 'Supination'),
        'FFTBAX': ('Plantarflexion', 'Dorsiflexion'),
        'FFTBAY': ('Abduction', 'Adduction'),
        'FFTBAZ': ('Pronation', 'Supination'),
        'HFTBAX': ('Plantarflexion', 'Dorsiflexion'),
        'HFTBAY': ('External', 'Internal'),
        'HFTBAZ': ('Eversion', 'Inversion'),
        'HFTFLX': ('Plantarflexion', 'Dorsiflexion'),
        'HFTFLY': ('External', 'Internal'),
        'HFTFLZ': ('Eversion', 'Inversion'),
        'HXFFAX': ('Plantarflexion', 'Dorsiflexion'),
        'HXFFAY': ('Valgus', 'Varus'),
        'HXFFAZ': ('NA', 'NA'),
        'TIBAX': ('Backward', 'Forward'),
        'TIBAY': ('Distal', 'Lateral'),
        'TIBAZ': ('External', 'Internal'),
    }
)


ofm.varlabels = _dict_with_side(ofm.varlabels_noside)
ofm.varnames = ofm.varlabels.keys()
ofm.varnames_noside = ofm.varlabels_noside.keys()

models_all.append(ofm)


#
# Plug-in Gait upper body kinematics
#
pig_upperbody = GaitModel()
pig_upperbody.desc = 'Plug-in Gait upper body kinematics'
pig_upperbody.read_strategy = 'split_xyz'

pig_upperbody.read_vars = _list_with_side(
    [
        'ThoraxAngles',
        'ShoulderAngles',
        'SpineAngles',
        'WristAngles',
        'ElbowAngles',
        'NeckAngles',
        'HeadAngles',
    ]
)

pig_upperbody.varlabels_noside = {
    'ThoraxAnglesX': 'Thorax flex/ext',
    'ThoraxAnglesY': 'Thorax lateral flex',
    'ThoraxAnglesZ': 'Thorax rotation',
    'ShoulderAnglesX': 'Shoulder flex/ext',
    'ShoulderAnglesY': 'Shoulder abd/add',
    'ShoulderAnglesZ': 'Shoulder rotation',
    'SpineAnglesX': 'Spine flex/ext',
    'SpineAnglesY': 'Spine lateral flex',
    'SpineAnglesZ': 'Spine rotation',
    'WristAnglesX': 'Wrist flex/ext',
    'WristAnglesY': 'Wrist ulnar/radial',
    'WristAnglesZ': 'Wrist sup/pron',
    'ElbowAnglesX': 'Elbow flex/ext',
    'ElbowAnglesY': 'Elbow y angle',
    'ElbowAnglesZ': 'Elbow rotation',
    'NeckAnglesX': 'Neck sagittal',
    'NeckAnglesY': 'Neck frontal',
    'NeckAnglesZ': 'Neck rotation',
    'HeadAnglesX': 'Head sagittal',
    'HeadAnglesY': 'Head frontal',
    'HeadAnglesZ': 'Head rotation',
}

# FIXME: for now, ylabel is just the degree sign for all vars
pig_upperbody.units = defaultdict(lambda: 'deg')
# FIXME: incomplete y descriptions
pig_upperbody.ydesc = defaultdict(lambda: ('', ''))  # FIXME: see Vicon docs
upperbody_ydesc_partial = _dict_with_side(
    {
        'ThoraxAnglesX': ('Backward', 'Forward'),
        'ThoraxAnglesY': ('Lateral', 'Medial'),
        'ThoraxAnglesZ': ('Anterior', 'Posterior'),
    }
)
pig_upperbody.ydesc.update(upperbody_ydesc_partial)
pig_upperbody.varlabels = _dict_with_side(pig_upperbody.varlabels_noside)
pig_upperbody.varnames = pig_upperbody.varlabels.keys()
pig_upperbody.varnames_noside = pig_upperbody.varlabels_noside.keys()

models_all.append(pig_upperbody)


#
# Plug-in Gait lower body kinematics
#
pig_lowerbody = GaitModel()
pig_lowerbody.desc = 'Plug-in Gait lower body kinematics'
pig_lowerbody.read_strategy = 'split_xyz'
pig_lowerbody.read_vars = _list_with_side(
    [
        'HipAngles',
        'KneeAngles',
        'AnkleAngles',
        'ForeFootAngles',
        'PelvisAngles',
        'FootProgressAngles',
    ]
)

pig_lowerbody.varlabels_noside = {
    'AnkleAnglesX': 'Ankle dorsi/plant',
    'AnkleAnglesY': 'Ankle adduction',
    'AnkleAnglesZ': 'Ankle rotation',
    'FootProgressAnglesZ': 'Foot progress angles',
    'ForeFootAnglesX': 'Fore/hindfoot dorsi/plant',
    'ForeFootAnglesY': 'Fore/hindfoot adduction',
    'ForeFootAnglesZ': 'Fore/hindfoot inversion',
    'HipAnglesX': 'Hip flexion',
    'HipAnglesY': 'Hip adduction',
    'HipAnglesZ': 'Hip rotation',
    'KneeAnglesX': 'Knee flexion',
    'KneeAnglesY': 'Knee adduction',
    'KneeAnglesZ': 'Knee rotation',
    'PelvisAnglesX': 'Pelvic tilt',
    'PelvisAnglesY': 'Pelvic obliquity',
    'PelvisAnglesZ': 'Pelvic rotation',
}

pig_lowerbody.varlabels = _dict_with_side(pig_lowerbody.varlabels_noside)

pig_lowerbody.varnames = pig_lowerbody.varlabels.keys()
pig_lowerbody.varnames_noside = pig_lowerbody.varlabels_noside.keys()

pig_lowerbody.gcd_normaldata_map = {
    'DorsiPlanFlex': 'AnkleAnglesX',
    'FootProgression': 'FootProgressAnglesZ',
    'FootRotation': 'AnkleAnglesZ',
    'HipAbAdduct': 'HipAnglesY',
    'HipFlexExt': 'HipAnglesX',
    'HipRotation': 'HipAnglesZ',
    'KneeFlexExt': 'KneeAnglesX',
    'KneeRotation': 'KneeAnglesZ',
    'KneeValgVar': 'KneeAnglesY',
    'PelvicObliquity': 'PelvisAnglesY',
    'PelvicRotation': 'PelvisAnglesZ',
    'PelvicTilt': 'PelvisAnglesX',
}

# unit is degrees for each kinematic variable
pig_lowerbody.units = defaultdict(lambda: 'deg')

pig_lowerbody.ydesc = _dict_with_side(
    {
        'AnkleAnglesX': ('Plantarflexion', 'Dorsiflexion'),
        'AnkleAnglesY': ('Abduction', 'Adduction'),
        'AnkleAnglesZ': ('External', 'Internal'),
        'FootProgressAnglesZ': ('External', 'Internal'),
        'ForeFootAnglesX': ('Plantarflexion', 'Dorsiflexion'),
        'ForeFootAnglesY': ('Abduction', 'Adduction'),
        'ForeFootAnglesZ': ('Eversion', 'Inversion'),
        'HipAnglesX': ('Extension', 'Flexion'),
        'HipAnglesY': ('Abduction', 'Adduction'),
        'HipAnglesZ': ('External', 'Internal'),
        'KneeAnglesX': ('Extension', 'Flexion'),
        'KneeAnglesY': ('Valgus', 'Varus'),
        'KneeAnglesZ': ('External', 'Internal'),
        'PelvisAnglesX': ('Posterior', 'Anterior'),
        'PelvisAnglesY': ('Down', 'Up'),
        'PelvisAnglesZ': ('External', 'Internal'),
    }
)

models_all.append(pig_lowerbody)


#
# Plug-in Gait lowerbody kinetics
#
pig_lowerbody_kinetics = GaitModel()
pig_lowerbody_kinetics.desc = 'Plug-in Gait lower body kinetics'
pig_lowerbody_kinetics.read_strategy = 'split_xyz'
pig_lowerbody_kinetics.read_vars = _list_with_side(
    [
        'HipMoment',
        'KneeMoment',
        'AnkleMoment',
        'HipPower',
        'KneePower',
        'AnklePower',
        'NormalisedGRF',
    ]
)

pig_lowerbody_kinetics.varlabels_noside = {
    'AnkleMomentX': 'Ankle dors/plan moment',
    'AnkleMomentY': 'Ankle ab/add moment',
    'AnkleMomentZ': 'Ankle rotation moment',
    'AnklePowerZ': 'Ankle power',
    'HipMomentX': 'Hip flex/ext moment',
    'HipMomentY': 'Hip ab/add moment',
    'HipMomentZ': 'Hip rotation moment',
    'HipPowerZ': 'Hip power',
    'KneeMomentX': 'Knee flex/ext moment',
    'KneeMomentY': 'Knee ab/add moment',
    'KneeMomentZ': 'Knee rotation moment',
    'KneePowerZ': 'Knee power',
    'NormalisedGRFX': 'Norm. GRF (x)',
    'NormalisedGRFY': 'Norm. GRF (y)',
    'NormalisedGRFZ': 'Norm. GRF (z)',
}

pig_lowerbody_kinetics.varlabels = _dict_with_side(
    pig_lowerbody_kinetics.varlabels_noside
)

pig_lowerbody_kinetics.varnames = pig_lowerbody_kinetics.varlabels.keys()
pig_lowerbody_kinetics.varnames_noside = pig_lowerbody_kinetics.varlabels_noside.keys()

pig_lowerbody_kinetics.gcd_normaldata_map = {
    'AnklePower': 'AnklePowerZ',
    'DorsiPlanFlexMoment': 'AnkleMomentX',
    'FootAbAdductMoment': 'AnkleMomentY',
    'FootRotationMoment': 'AnkleMomentZ',
    'HipAbAdductMoment': 'HipMomentY',
    'HipFlexExtMoment': 'HipMomentX',
    'HipPower': 'HipPowerZ',
    'HipRotationMoment': 'HipMomentZ',
    'KneeFlexExtMoment': 'KneeMomentX',
    'KneePower': 'KneePowerZ',
    'KneeRotationMoment': 'KneeMomentZ',
    'KneeValgVarMoment': 'KneeMomentY',
}

pig_lowerbody_kinetics.units = _dict_with_side(
    {
        'AnkleMomentX': 'Nm/kg',
        'AnkleMomentY': 'Nm/kg',
        'AnkleMomentZ': 'Nm/kg',
        'AnklePowerZ': 'W/kg',
        'HipMomentX': 'Nm/kg',
        'HipMomentY': 'Nm/kg',
        'HipMomentZ': 'Nm/kg',
        'HipPowerZ': 'W/kg',
        'KneeMomentX': 'Nm/kg',
        'KneeMomentY': 'Nm/kg',
        'KneeMomentZ': 'Nm/kg',
        'KneePowerZ': 'W/kg',
        'NormalisedGRFX': '%BW',
        'NormalisedGRFY': '%BW',
        'NormalisedGRFZ': '%BW',
    }
)

pig_lowerbody_kinetics.ydesc = _dict_with_side(
    {
        'AnkleMomentX': ('Dorsiflexion', 'Plantarflexion'),
        'AnkleMomentY': ('Adduction', 'Abduction'),
        'AnkleMomentZ': ('Internal', 'External'),  # FIXME: check
        'AnklePowerZ': ('Absorbing', 'Generating'),
        'HipMomentX': ('Flexion', 'Extension'),
        'HipMomentY': ('Adduction', 'Abduction'),
        'HipMomentZ': ('Internal', 'External'),
        'HipPowerZ': ('Absorbing', 'Generating'),
        'KneeMomentX': ('Flexion', 'Extension'),
        'KneeMomentY': ('Varus', 'Valgus'),
        'KneeMomentZ': ('Internal', 'External'),
        'KneePowerZ': ('Absorbing', 'Generating'),
        'NormalisedGRFX': ('', ''),
        'NormalisedGRFY': ('', ''),
        'NormalisedGRFZ': ('Downward', 'Upward'),
    }
)

pig_lowerbody_kinetics.is_kinetic_var = lambda varname: True

models_all.append(pig_lowerbody_kinetics)

#
# Muscle length (MuscleLength.mod)
#

musclelen = GaitModel()
musclelen.desc = 'Muscle length (MuscleLength.mod)'
musclelen.read_strategy = 'last'

musclelen.varlabels_noside = {
    'AdBrLength': 'AdBrLength',
    'AdLoLength': 'AdLoLength',
    'AdMaInfLength': 'AdMaInfLength',
    'AdMaMidLength': 'AdMaMidLength',
    'AdMaSupLength': 'AdMaSupLength',
    'BiFLLength': 'Biceps femoris length',
    'BiFSLength': 'BiFSLength',
    'ExDLLength': 'ExDLLength',
    'ExHLLength': 'ExHLLength',
    'FlDLLength': 'FlDLLength',
    'FlHLLength': 'FlHLLength',
    'GMedAntLength': 'GMedAntLength',
    'GMedMidLength': 'GMedMidLength',
    'GMedPosLength': 'GMedPosLength',
    'GMinAntLength': 'GMinAntLength',
    'GMinMidLength': 'GMinMidLength',
    'GMinPosLength': 'GMinPosLength',
    'GemeLength': 'GemeLength',
    'GlMaInfLength': 'GlMaInfLength',
    'GlMaMidLength': 'GlMaMidLength',
    'GlMaSupLength': 'GlMaSupLength',
    'GracLength': 'Gracilis length',
    'IliaLength': 'IliaLength',
    'LaGaLength': 'Lateral gastrocnemius length',
    'MeGaLength': 'Medial gastrocnemius length',
    'PELOLength': 'PELOLength',
    'PeBrLength': 'PeBrLength',
    'PeTeLength': 'PeTeLength',
    'PectLength': 'PectLength',
    'PeriLength': 'PeriLength',
    'PsoaLength': 'Psoas length',
    'QuFeLength': 'QuFeLength',
    'ReFeLength': 'Rectus femoris length',
    'SartLength': 'SartLength',
    'SeMeLength': 'Semimembranosus length',
    'SeTeLength': 'Semitendinosus length',
    'SoleLength': 'Soleus length',
    'TiAnLength': 'Tibialis anterior length',
    'TiPoLength': 'TiPoLength',
    'VaInLength': 'VaInLength',
    'VaLaLength': 'VaLaLength',
    'VaMeLength': 'VaMeLength',
}

musclelen.varlabels = _dict_with_side(musclelen.varlabels_noside)
musclelen.read_vars = musclelen.varlabels.keys()
musclelen.varnames = musclelen.read_vars
musclelen.varnames_noside = musclelen.varlabels_noside.keys()

musclelen.units = defaultdict(lambda: 'Norm length')  # FIXME: % of what?
musclelen.ydesc = defaultdict(lambda: ('', ''))

models_all.append(musclelen)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Time-distance computations

@author: Jussi (jnu@iki.fi)
"""

from __future__ import division
import logging
import numpy as np
from collections import OrderedDict, defaultdict

from .envutils import GaitDataError
from . import c3d

logger = logging.getLogger(__name__)


def group_analysis(an_list, fun=np.mean):
    """Apply function (e.g. mean or stddev) to analysis dicts.

    Parameters
    ----------
    an_list : list
        List of analysis dicts returned by read_data.get_analysis(). All dicts
        must have the same condition label.
    fun : function
        The reducing function to apply, by default np.mean. This must accept a
        1-D ndarray of values and return a single value. Examples of useful
        functions would be np.mean, np.std and np.median.

    Returns
    -------
    dict
        The resulting analysis dict after the reducing function has been
        applied. The condition label is identical with the input dicts.
    """

    if not isinstance(an_list, list):
        raise TypeError('Need a list of analysis dicts')
    if not an_list:
        return None

    # check conditions
    condsets = [set(an.keys()) for an in an_list]
    conds = condsets[0]
    if not all(cset == conds for cset in condsets):
        raise RuntimeError('Conditions need to match between analysis dicts')

    # figure out variables that are in all of the analysis dicts
    for cond in conds:
        varsets = [set(an[cond].keys()) for an in an_list for cond in conds]
    vars_ = set.intersection(*varsets)
    not_in_all = set.union(*varsets) - vars_
    if not_in_all:
        logger.warning(
            'Some analysis dicts are missing the following variables: %s'
            % ' '.join(not_in_all)
        )

    # gather data and apply function
    res = defaultdict(lambda: defaultdict(dict))
    for cond in conds:
        for var in vars_:
            # this will fail if vars are not strictly matched between dicts
            res[cond][var]['unit'] = an_list[0][cond][var]['unit']
            for context in ['Right', 'Left']:
                # gather valus from analysis dicts
                allvals = np.array(
                    [
                        an[cond][var][context]
                        for an in an_list
                        if context in an[cond][var]
                    ]
                )
                # filter out missing values (nans)
                allvals = allvals[~np.isnan(allvals)]
                res[cond][var][context] = fun(allvals) if allvals.size else np.nan
    return res


def _group_analysis_trials(trials):
    """Multitrial analysis from given trials (.c3d files).
    trials: dict of lists keyed by condition name
    If there are multiple trials per condition, they will be averaged.
    """
    res_avg_all = OrderedDict()  # preserve condition ordering
    res_std_all = OrderedDict()  # for plots etc.
    for cond_label, cond_files in trials.items():
        ans = list()
        for c3dfile in cond_files:
            try:
                an = c3d.get_analysis(c3dfile, condition=cond_label)
                ans.append(an)
            except GaitDataError:
                logger.warning('no analysis values found in %s' % c3dfile)
        if ans:
            res_avg = group_analysis(ans)
            res_std = group_analysis(ans, fun=np.std)
            res_avg_all.update(res_avg)
            res_std_all.update(res_std)
    return res_avg_all, res_std_all


def _pick_common_vars(values, vars_wanted=None):
    """Helper to pick analysis vars data that exist for
    all conditions. Returns vars and corresponding units"""
    conds = list(values.keys())
    vals_1 = values[conds[0]]
    varsets = [set(values[cond].keys()) for cond in conds]
    # vars common to all conditions
    vars_common = set.intersection(*varsets)
    if vars_wanted is not None:
        # pick specified vars that appear in all of the conditions
        vars_wanted_set = set(vars_wanted)
        vars_ok = set.intersection(vars_wanted_set, vars_common)
        if vars_wanted_set - vars_ok:
            logger.warning(
                'some conditions are missing variables: %s'
                % (vars_wanted_set - vars_ok)
            )
        # preserve original var order
        vars_ = [var for var in vars_wanted if var in vars_ok]
    else:
        vars_ = list(vars_common)
    units = [vals_1[var]['unit'] for var in vars_]
    return conds, vars_, units



# -*- coding: utf-8 -*-
"""
gaitutils custom Qt widgets

@author: Jussi (jnu@iki.fi)
"""

from PyQt5 import QtCore, QtWidgets
from PyQt5.QtCore import QObject, pyqtSignal
import logging


class NiceListWidgetItem(QtWidgets.QListWidgetItem):
    """Make list items more pythonic"""

    def __init__(self, *args, **kwargs):
        # don't pass this arg to superclass __init__
        checkable = kwargs.pop('checkable')
        super(NiceListWidgetItem, self).__init__(*args, **kwargs)
        if checkable:
            self.setFlags(self.flags() | QtCore.Qt.ItemIsUserCheckable)

    @property
    def text(self):
        return super(NiceListWidgetItem, self).text()

    @property
    def userdata(self):
        return super(NiceListWidgetItem, self).data(QtCore.Qt.UserRole)

    @userdata.setter
    def userdata(self, _data):
        if _data is not None:
            super(NiceListWidgetItem, self).setData(QtCore.Qt.UserRole, _data)

    @property
    def checkstate(self):
        return super(NiceListWidgetItem, self).checkState()

    @checkstate.setter
    def checkstate(self, checked):
        state = QtCore.Qt.Checked if checked else QtCore.Qt.Unchecked
        super(NiceListWidgetItem, self).setCheckState(state)


class NiceListWidget(QtWidgets.QListWidget):
    """Adds some conveniences to QListWidget"""

    def __init__(self, parent=None):
        super(NiceListWidget, self).__init__(parent)

    @property
    def items(self):
        """Yield all list items.

        NB: be careful when modifying list items in a loop with the generator -
        count() is evaluated only once, so the generator may return items that
        have already been deleted"""
        for i in range(self.count()):
            yield self.item(i)

    @property
    def checked_items(self):
        """Yield checked items"""
        return (item for item in self.items if item.checkstate)

    def check_all(self):
        """Check all items"""
        for item in self.items:
            item.checkstate = True

    def check_none(self):
        """Uncheck all items"""
        for item in self.items:
            item.checkstate = False

    def add_item(self, txt, data=None, checkable=False, checked=False):
        """Add checkable item with data. Select new item."""
        item = NiceListWidgetItem(txt, self, checkable=checkable)
        item.userdata = data
        if checkable:
            item.checkstate = checked
        self.setCurrentItem(item)

    def rm_current_item(self):
        """Remove currently selected item"""
        return self.takeItem(self.row(self.currentItem()))


class QtHandler(logging.Handler):
    """Qt logging handler"""

    def __init__(self):
        logging.Handler.__init__(self)

    def emit(self, record):
        record = self.format(record)
        if record:
            XStream.stdout().write('%s\n' % record)


class XStream(QtCore.QObject):
    """Stream for Qt logging handler"""

    _stdout = None
    _stderr = None
    messageWritten = QtCore.pyqtSignal(str)

    def flush(self):
        pass

    def fileno(self):
        return -1

    def write(self, msg):
        if not self.signalsBlocked():
            self.messageWritten.emit(msg)

    @staticmethod
    def stdout():
        if not XStream._stdout:
            XStream._stdout = XStream()
            # sys.stdout = XStream._stdout  # also capture stdout
        return XStream._stdout

    @staticmethod
    def stderr():
        if not XStream._stderr:
            XStream._stderr = XStream()
            # sys.stderr = XStream._stderr  # ... and stderr
        return XStream._stderr


class ProgressBar(QtWidgets.QProgressDialog):
    """Qt progress bar with reasonable defaults"""

    # custom signal to indicate that the operation was canceled
    _canceled = pyqtSignal()

    def __init__(self, title):
        super(self.__class__, self).__init__()
        self.setWindowTitle(title)
        self.cancelbutton = QtWidgets.QPushButton('Cancel')
        self.setCancelButton(self.cancelbutton)
        # set a custom cancel signal, since the default one immediately
        # close the progress bar
        self.cancelbutton.disconnect()
        self.cancelbutton.clicked.connect(self._cancel)
        self.setMinimum(0)
        self.setMaximum(100)
        self.setGeometry(500, 300, 500, 100)
        # self.setAutoClose(False)
        self.setAutoReset(False)
        self.show()

    def _cancel(self):
        """Custom cancel handler"""
        self.cancelbutton.setText('Wait...')
        self.cancelbutton.setEnabled(False)
        QtWidgets.QApplication.processEvents()
        self._canceled.emit()

    def update(self, text, p):
        """Update bar, showing text and bar at p%"""
        self.setLabelText(text)
        self.setValue(p)
        # update right away in case that thread is blocked
        QtWidgets.QApplication.processEvents()


class ProgressSignals(QObject):
    """Used to emit progress signals across threads"""

    progress = pyqtSignal(object, object)

    def __init__(self):
        super(ProgressSignals, self).__init__()
        # this flag can be checked to see whether the operation was canceled
        self.canceled = False

    def cancel(self):
        """Raise the cancel flag"""
        self.canceled = True
